import multiprocessing

from multiprocessing import JoinableQueue

from ms_deisotope.processor import MSFileLoader

from ms_deisotope.feature_map.quick_index import index as build_scan_index
from ms_deisotope.task import TaskBase

from .collator import ScanCollator
from .process import ScanIDYieldingProcess, ScanTransformingProcess


class ScanGeneratorBase(object):

    def configure_iteration(self, start_scan=None, end_scan=None, max_scans=None):
        raise NotImplementedError()

    def make_iterator(self, start_scan=None, end_scan=None, max_scans=None):
        raise NotImplementedError()

    _iterator = None

    def __iter__(self):
        return self

    def __next__(self):
        if self._iterator is None:
            self._iterator = self.make_iterator()
        return next(self._iterator)

    def next(self):
        return self.__next__()

    def close(self):
        pass

    @property
    def scan_source(self):
        return None

    _deconvoluting = False

    @property
    def deconvoluting(self):
        return self._deconvoluting

    @deconvoluting.setter
    def deconvoluting(self, value):
        self._deconvoluting = value

    _ms1_averaging = 0

    @property
    def ms1_averaging(self):
        return self._ms1_averaging

    @ms1_averaging.setter
    def ms1_averaging(self, value):
        self._ms1_averaging = value

    _ignore_tandem_scans = False

    @property
    def ignore_tandem_scans(self):
        return self._ignore_tandem_scans

    @ignore_tandem_scans.setter
    def ignore_tandem_scans(self, value):
        self._ignore_tandem_scans = value

    _extract_only_tandem_envelopes = False

    @property
    def extract_only_tandem_envelopes(self):
        return self._extract_only_tandem_envelopes

    @extract_only_tandem_envelopes.setter
    def extract_only_tandem_envelopes(self, value):
        self._extract_only_tandem_envelopes = value


class ScanGenerator(TaskBase, ScanGeneratorBase):
    def __init__(self, ms_file, number_of_helpers=4,
                 ms1_peak_picking_args=None, msn_peak_picking_args=None,
                 ms1_deconvolution_args=None, msn_deconvolution_args=None,
                 extract_only_tandem_envelopes=False, ignore_tandem_scans=False,
                 ms1_averaging=0, deconvolute=True, verbose=False):
        self.ms_file = ms_file
        self.ignore_tandem_scans = ignore_tandem_scans

        self.scan_ids_exhausted_event = multiprocessing.Event()

        self._iterator = None

        self._scan_yielder_process = None
        self._deconv_process = None

        self._input_queue = None
        self._output_queue = None
        self._deconv_helpers = None
        self._order_manager = None

        self.number_of_helpers = number_of_helpers

        self.ms1_peak_picking_args = ms1_peak_picking_args
        self.msn_peak_picking_args = msn_peak_picking_args
        self.ms1_averaging = ms1_averaging

        self.deconvoluting = deconvolute
        self.ms1_deconvolution_args = ms1_deconvolution_args
        self.msn_deconvolution_args = msn_deconvolution_args
        self.extract_only_tandem_envelopes = extract_only_tandem_envelopes
        self._scan_interval_tree = None
        self.verbose = verbose
        self.log_controller = self.ipc_logger()

    @property
    def scan_source(self):
        return self.ms_file

    def join(self):
        if self._scan_yielder_process is not None:
            self._scan_yielder_process.join()
        if self._deconv_process is not None:
            self._deconv_process.join()
        if self._deconv_helpers is not None:
            for helper in self._deconv_helpers:
                helper.join()

    def _terminate(self):
        if self._scan_yielder_process is not None:
            self._scan_yielder_process.terminate()
        if self._deconv_process is not None:
            self._deconv_process.terminate()
        if self._deconv_helpers is not None:
            for helper in self._deconv_helpers:
                helper.terminate()

    def _preindex_file(self):
        reader = MSFileLoader(self.ms_file, use_index=False)
        try:
            reader.prebuild_byte_offset_file(self.ms_file)
        except AttributeError:
            # the type does not support this type of indexing
            pass
        except IOError:
            # the file could not be written
            pass
        except Exception as e:
            # something else went wrong
            self.error("An error occurred while pre-indexing.", e)

    def _make_interval_tree(self, start_scan, end_scan):
        reader = MSFileLoader(self.ms_file)
        if start_scan is not None:
            start_ix = reader.get_scan_by_id(start_scan).index
        else:
            start_ix = 0
        if end_scan is not None:
            end_ix = reader.get_scan_by_id(end_scan).index
        else:
            end_ix = len(reader)
        reader.reset()
        index, interval_tree = build_scan_index(
            reader, self.number_of_helpers + 1, (start_ix, end_ix))
        self._scan_interval_tree = interval_tree

    def _make_transforming_process(self):
        return ScanTransformingProcess(
            self.ms_file,
            self._input_queue,
            self._output_queue,
            self.scan_ids_exhausted_event,
            ms1_peak_picking_args=self.ms1_peak_picking_args,
            msn_peak_picking_args=self.msn_peak_picking_args,
            ms1_deconvolution_args=self.ms1_deconvolution_args,
            msn_deconvolution_args=self.msn_deconvolution_args,
            envelope_selector=self._scan_interval_tree,
            log_handler=self.log_controller.sender(),
            ms1_averaging=self.ms1_averaging,
            deconvolute=self.deconvoluting,
            verbose=self.verbose)

    def _make_collator(self):
        return ScanCollator(
            self._output_queue, self.scan_ids_exhausted_event, self._deconv_helpers,
            self._deconv_process, input_queue=self._input_queue,
            include_fitted=not self.deconvoluting)

    def _initialize_workers(self, start_scan=None, end_scan=None, max_scans=None):
        try:
            self._input_queue = JoinableQueue(int(1e6))
            self._output_queue = JoinableQueue(5000)
        except OSError:
            # Not all platforms permit limiting the size of queues
            self._input_queue = JoinableQueue()
            self._output_queue = JoinableQueue()

        self._preindex_file()

        if self.extract_only_tandem_envelopes:
            self.log("Constructing Scan Interval Tree")
            self._make_interval_tree(start_scan, end_scan)

        self._terminate()
        self._scan_yielder_process = ScanIDYieldingProcess(
            self.ms_file, self._input_queue, start_scan=start_scan, end_scan=end_scan,
            max_scans=max_scans, no_more_event=self.scan_ids_exhausted_event,
            ignore_tandem_scans=self.ignore_tandem_scans, batch_size=1)
        self._scan_yielder_process.start()

        self._deconv_process = self._make_transforming_process()

        self._deconv_helpers = []

        for _ in range(self.number_of_helpers):
            self._deconv_helpers.append(self._make_transforming_process())
        self._deconv_process.start()

        self._order_manager = self._make_collator()

    def make_iterator(self, start_scan=None, end_scan=None, max_scans=None):
        self._initialize_workers(start_scan, end_scan, max_scans)

        for scan in self._order_manager:
            yield scan
        self.log_controller.stop()
        self.join()
        self._terminate()

    def configure_iteration(self, start_scan=None, end_scan=None, max_scans=None):
        self._iterator = self.make_iterator(start_scan, end_scan, max_scans)

    def close(self):
        self._terminate()



import json

from collections import namedtuple, deque

from ms_deisotope.peak_dependency_network import Interval, IntervalTreeNode


class BoundingBox(namedtuple("BoundingBox", ['mz', 'rt'])):
    def merge(self, other):
        min_mz = min(self[0].start, other[0].start)
        max_mz = max(self[0].end, other[0].end)
        min_rt = min(self[1].start, other[1].start)
        max_rt = max(self[1].end, other[1].end)
        return self.__class__(Interval(min_mz, max_mz), Interval(min_rt, max_rt))

    def overlaps(self, other):
        return self.mz.overlaps(other.mz) and self.rt.overlaps(other.rt)

    def __add__(self, other):
        return self.merge(other)


def extract_intervals(scan_iterator, time_radius=5., mz_lower=2., mz_higher=3.):
    intervals = []
    for scan, products in scan_iterator:
        for product in products:
            intervals.append(BoundingBox(
                Interval(max(0, product.precursor_information.mz - mz_lower),
                         product.precursor_information.mz + mz_higher),
                Interval(max(0, product.scan_time - time_radius),
                         product.scan_time + time_radius)))
    return intervals


def nest_2d_intervals(intervals):
    out = []
    for interval in intervals:
        mz, rt = interval
        rt.members.append(mz)
        out.append(rt)
    return out


def make_rt_tree(intervals):
    nested = nest_2d_intervals(intervals)
    tree = IntervalTreeNode.build(nested)
    return tree


class ScanIntervalTree(object):
    @classmethod
    def build(cls, scan_iterator, time_radius=3., mz_lower=2., mz_higher=3.):
        intervals = extract_intervals(
            scan_iterator, time_radius=time_radius,
            mz_lower=mz_lower, mz_higher=mz_higher)
        return cls(make_rt_tree(intervals), intervals)

    def __init__(self, rt_tree, original_intervals=None):
        self.rt_tree = rt_tree
        self.original_intervals = original_intervals

    def get_mz_intervals_for_rt(self, rt_point):
        if self.rt_tree is None:
            return [[0, float('inf')]]
        intervals = [
            i.members[0] for i in self.rt_tree.contains_point(rt_point)
        ]
        intervals = sorted([[i.start, i.end] for i in intervals])
        out = []
        if len(intervals) == 0:
            return []
        last = intervals[0]
        for interval in intervals[1:]:
            if interval[0] < last[1]:
                last[1] = interval[1]
            else:
                out.append(last)
                last = interval
        out.append(last)
        return out

    def __call__(self, scan):
        intervals = self.get_mz_intervals_for_rt(scan.scan_time)
        return intervals

    def serialize(self, handle):
        work_queue = deque()
        id_counter = 0
        node_id_map = {}

        def serialize_contained(interval_list):
            out = []
            for interval in interval_list:
                item = {"start": interval.start, "end": interval.end}
                inner = [{"start": i.start, "end": i.end}
                         for i in interval.members]
                item['members'] = inner
                out.append(item)
            return out

        root = self.rt_tree
        node_dict = {
            "id": id_counter,
            "contained": serialize_contained(root.contained),
            "center": root.center,
            "start": root.start,
            "end": root.end,
            "left": None,
            "right": None
        }
        root_id = node_dict['id']
        node_id_map[node_dict['id']] = node_dict
        id_counter += 1
        work_queue.append((root.left, node_dict['id'], 'left'))
        work_queue.append((root.right, node_dict['id'], 'right'))

        while work_queue:
            node, parent_id, side = work_queue.popleft()
            if node is None:
                continue
            node_dict = {
                "id": id_counter,
                "contained": serialize_contained(node.contained),
                "center": node.center,
                "start": node.start,
                "end": node.end,
                "left": None,
                "right": None
            }
            node_id_map[node_dict['id']] = node_dict
            id_counter += 1
            parent = node_id_map[parent_id]
            parent[side] = node_dict
            work_queue.append((node.left, node_dict['id'], 'left'))
            work_queue.append((node.right, node_dict['id'], 'right'))

        json.dump(node_id_map[root_id], handle)

    @classmethod
    def load(cls, handle):
        data = json.load(handle)

        def load_contained(interval_list):
            out = []
            for item in interval_list:
                inner = [Interval(**i) for i in item['members']]
                out.append(Interval(start=item['start'], end=item['end'], members=inner))
            return out

        def load_node_single(node_dict):
            contained = node_dict.get("contained", [])
            contained = load_contained(contained)
            node = IntervalTreeNode(
                node_dict['center'], left=None, contained=contained,
                right=None)
            node.start = node_dict['start']
            node.end = node_dict['end']
            return node
        root = load_node_single(data)

        work_queue = deque()

        work_queue.append((data['left'], root, 'left'))
        work_queue.append((data['right'], root, 'right'))

        while work_queue:
            node_dict, parent, side = work_queue.popleft()
            if node_dict is None:
                continue
            node = load_node_single(node_dict)
            node.parent = parent
            node.level = parent.level + 1
            if side == 'left':
                parent.left = node
            elif side == 'right':
                parent.right = node
            else:
                raise ValueError(side)
            work_queue.append((node_dict['left'], node, 'left'))
            work_queue.append((node_dict['right'], node, 'right'))
        return cls(root)



'''A collection of common base classes for types that
load data for :class:`~.Scan` objects.
'''
import abc
from weakref import WeakValueDictionary

from ms_deisotope.utils import add_metaclass

from ms_deisotope.data_source.metadata.file_information import FileInformation


from .scan import Scan
from .scan_iterator import (
    _SingleScanIteratorImpl,
    _GroupedScanIteratorImpl,
    _FakeGroupedScanIteratorImpl)


@add_metaclass(abc.ABCMeta)
class ScanDataSource(object):
    """An Abstract Base Class describing an object
    which can provide a consistent set of accessors
    for a particular format of mass spectrometry data.

    Data files come in many shapes and sizes, with different
    underlying structures. This class provides an API that
    should make features as consistent as possible to clients
    of :class:`Scan` objects.
    """

    def _make_scan(self, data):
        return Scan(data, self)

    def _pick_peaks_vendor(self, scan, *args, **kwargs):
        """Invoke the underlying data access library's peak picking procedure.

        Not available for open format readers, where behavior will default to
        the :mod:`ms_peak_picker` algorithm.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        peak_set: :class:`ms_peak_picker.PeakIndex`

        Raises
        ------
        NotImplementedError:
            When there is no method available for the given scan and/or data source
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _scan_arrays(self, scan):
        """Returns raw data arrays for m/z and intensity

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        mz: np.array
            An array of m/z values for this scan
        intensity: np.array
            An array of intensity values for this scan
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _precursor_information(self, scan):
        """Returns information about the precursor ion,
        if any, that this scan was derived form.

        Returns `None` if this scan has no precursor ion

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        PrecursorInformation
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _scan_title(self, scan):
        """Returns a verbose name for this scan, if one
        was stored in the file. Usually includes both the
        scan's id string, as well as information about the
        original file and format.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        str
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _scan_id(self, scan):
        """Returns the scan's id string, a unique
        identifier for this scan in the context of
        the data file it is recordered in

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        str
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _scan_index(self, scan):
        """Returns the base 0 offset from the start
        of the data file in number of scans to reach
        this scan.

        If the original format does not natively include
        an index value, this value may be computed from
        the byte offset index.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _ms_level(self, scan):
        """Returns the degree of exponential fragmentation
        used to produce this scan. 1 refers to a survey scan
        of unfragmented ions, 2 refers to a tandem scan derived
        from an ms level 1 ion, and so on.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _scan_time(self, scan):
        """Returns the time in minutes from the start of data
        acquisition to when this scan was acquired.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        float
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _is_profile(self, scan):
        """Returns whether the scan contains profile data (`True`)
        or centroided data (`False`).

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        bool
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _polarity(self, scan):
        """Returns whether this scan was acquired in positive mode (+1)
        or negative mode (-1).

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _activation(self, scan):
        """Returns information about the activation method used to
        produce this scan, if any.

        Returns :const:`None` for MS1 scans

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a :class:`dict`

        Returns
        -------
        ActivationInformation
        """
        raise NotImplementedError()

    def _acquisition_information(self, scan):
        return None

    def _isolation_window(self, scan):
        return None

    def _instrument_configuration(self, scan):
        return None

    def _annotations(self, scan):
        return dict()


@add_metaclass(abc.ABCMeta)
class ScanIterator(ScanDataSource):
    """An Abstract Base Class that extends ScanDataSource
    with additional requirements that enable clients of the
    class to treat the object as an iterator over the underlying
    data file.
    """

    iteration_mode = 'group'

    def has_ms1_scans(self):
        '''Checks if this :class:`ScanDataSource` contains MS1 spectra.

        Returns
        -------
        :class:`bool` or :const:`None`
            Returns a boolean value if the presence of MS1 scans is known for certain, or :const:`None`
            if it cannot be determined in the case of missing metadata.
        '''
        return True

    def has_msn_scans(self):
        '''Checks if this :class:`ScanDataSource` contains MSn spectra.

        Returns
        -------
        :class:`bool` or :const:`None`
            Returns a boolean value if the presence of MSn scans is known for certain, or :const:`None`
            if it cannot be determined in the case of missing metadata.
        '''
        return True

    @abc.abstractmethod
    def next(self):
        '''Advance the iterator, fetching the next :class:`~.ScanBunch` or :class:`~.ScanBase`
        depending upon iteration strategy.

        Returns
        -------
        :class:`~.ScanBunch` or :class:`~.ScanBase`
        '''
        raise NotImplementedError()

    def __next__(self):
        '''Advance the iterator, fetching the next :class:`~.ScanBunch` or :class:`~.ScanBase`
        depending upon iteration strategy.

        Returns
        -------
        :class:`~.ScanBunch` or :class:`~.ScanBase`
        '''
        return self.next()

    def __iter__(self):
        return self

    def reset(self):
        '''Reset the iterator, if possible, and clear any caches.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def _make_default_iterator(self):
        '''Set up the default iterator for the :class:`ScanIterator`.
        '''
        raise NotImplementedError()

    def make_iterator(self, iterator=None, grouped=None):
        """Configure the :class:`ScanIterator`'s behavior, selecting it's iteration strategy over
        either its default iterator or the provided ``iterator`` argument.

        Parameters
        ----------
        iterator : Iterator, optional
            The iterator to manipulate. If missing, the default
            iterator will be used.
        grouped : bool, optional
            Whether the iterator should be grouped and produce :class:`.ScanBunch` objects
            or single :class:`.Scan`. If :const:`None` is passed, :meth:`has_ms1_scans` will be
            be used instead. Defaults to :const:`None`.
        """
        if grouped is None:
            grouped = self.has_ms1_scans()

        if grouped:
            self._producer = self._scan_group_iterator(iterator)
            self.iteration_mode = 'group'
        else:
            self._producer = self._single_scan_iterator(iterator)
            self.iteration_mode = 'single'
        return self

    def _make_cache_key(self, scan):
        return scan.id

    def _cache_scan(self, scan):
        key = self._make_cache_key(scan)
        self._scan_cache[key] = scan
        return key

    def _validate(self, scan):
        return True

    def _single_scan_iterator(self, iterator=None):
        if iterator is None:
            iterator = self._make_default_iterator()

        impl = _SingleScanIteratorImpl.from_scan_source(iterator, self)
        return impl

    def _scan_group_iterator(self, iterator=None):
        if iterator is None:
            iterator = self._make_default_iterator()
        if self.has_ms1_scans():
            impl = _GroupedScanIteratorImpl.from_scan_source(iterator, self)
        else:
            impl = _FakeGroupedScanIteratorImpl.from_scan_source(iterator, self)
        return impl

    def _scan_cleared(self, scan):
        self.scan_cache.pop(self._make_cache_key(scan), None)

    def initialize_scan_cache(self):
        '''Initialize a cache which keeps track of which :class:`~.Scan`
        objects are still in memory using a :class:`weakref.WeakValueDictionary`.

        When a scan is requested, if the scan object is found in the cahce, the
        existing object is returned rather than re-read from disk.
        '''
        self._scan_cache = WeakValueDictionary()

    @property
    def scan_cache(self):
        '''A :class:`weakref.WeakValueDictionary` mapping used to retrieve
        scans from memory if available before re-reading them from disk.
        '''
        return self._scan_cache

    @scan_cache.setter
    def scan_cache(self, value):
        self._scan_cache = value


@add_metaclass(abc.ABCMeta)
class RandomAccessScanSource(ScanIterator):
    """An Abstract Base Class that extends ScanIterator
    with additional requirements that the implementation support
    random access to individual scans. This should be doable by unique
    identifier, sequential index, or by scan time.
    """

    @abc.abstractmethod
    def get_scan_by_id(self, scan_id):
        """Retrieve the scan object for the specified scan id.

        If the scan object is still bound and in memory somewhere,
        a reference to that same object will be returned. Otherwise,
        a new object will be created.

        Parameters
        ----------
        scan_id : str
            The unique scan id value to be retrieved

        Returns
        -------
        Scan
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_scan_by_time(self, time):
        """Retrieve the scan object for the specified scan time.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        time : float
            The time to get the nearest scan from

        Returns
        -------
        Scan
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_scan_by_index(self, index):
        """Retrieve the scan object for the specified scan index.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        index: int
            The index to get the scan for

        Returns
        -------
        Scan
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        '''Reconstruct an iterator which will start from the scan matching one of ``scan_id``,
        ``rt``, or ``index``. Only one may be provided.

        After invoking this method, the iterator this object wraps will be changed to begin
        yielding scan bunchs (or single scans if ``grouped`` is ``False``).

        This method will trigger several random-access operations, making it prohibitively
        expensive for normally compressed files.

        Arguments
        ---------
        scan_id: str, optional
            Start from the scan with the specified id.
        rt: float, optional
            Start from the scan nearest to specified time (in minutes) in the run. If no
            exact match is found, the nearest scan time will be found, rounded up.
        index: int, optional
            Start from the scan with the specified index.
        require_ms1: bool, optional
            Whether the iterator must start from an MS1 scan. True by default.
        grouped: bool, optional
            whether the iterator should yield scan bunches or single scans. True by default.
        '''
        raise NotImplementedError()

    def _locate_ms1_scan(self, scan, search_range=150):
        i = 0
        initial_scan = scan
        if (self.has_ms1_scans() is False):
            raise IndexError('Cannot locate MS1 Scan')
        while scan.ms_level != 1 and i < search_range:
            i += 1
            if scan.index <= 0:
                break
            scan = self.get_scan_by_index(scan.index - 1)
        if scan.ms_level == 1:
            return scan
        scan = initial_scan
        i = 0
        while scan.ms_level != 1 and i < search_range:
            i += 1
            try:
                scan = self.get_scan_by_index(scan.index + 1)
            except IndexError:
                raise IndexError("Cannot locate MS1 Scan")
        return scan

    def find_previous_ms1(self, start_index):
        '''Locate the MS1 scan preceding ``start_index``, iterating backwards through
        scans until either the first scan is reached or an MS1 scan is found.

        Returns
        -------
        :class:`~.ScanBase` or :const:`None` if not found
        '''
        if self.has_ms1_scans() is False:
            return None
        index = start_index - 1
        while index >= 0:
            try:
                scan = self.get_scan_by_index(index)
                if scan.ms_level == 1:
                    return scan
                index -= 1
            except (IndexError, KeyError):
                return None
        return None

    def find_next_ms1(self, start_index):
        '''Locate the MS1 scan following ``start_index``, iterating forwards through
        scans until either the last scan is reached or an MS1 scan is found.

        Returns
        -------
        :class:`~.ScanBase` or :const:`None` if not found
        '''
        if self.has_ms1_scans() is False:
            return None
        index = start_index + 1
        n = len(self.index)
        while index < n:
            try:
                scan = self.get_scan_by_index(index)
                if scan.ms_level == 1:
                    return scan
                index += 1
            except (IndexError, KeyError):
                return None
        return None

    @abc.abstractmethod
    def __len__(self):
        raise NotImplementedError()

    def __getitem__(self, i):
        """Retrieve the scan object for the specified scan index.

        This internally calls :meth:`get_scan_by_index` but supports
        slicing and negative indexing.

        Parameters
        ----------
        index: int
            The index to get the scan for

        Returns
        -------
        Scan
        """
        if isinstance(i, slice):
            n = len(self)
            scans = []
            start, stop, step = i.indices(n)
            for i in range(start, stop, step):
                scans.append(self[i])
            return scans
        elif i < 0:
            n = len(self)
            i = n + i
        return self.get_scan_by_index(i)


@add_metaclass(abc.ABCMeta)
class ScanFileMetadataBase(object):
    """Objects implementing this interface can describe the original source
    files, instrument configuration, and data processing parameters used to
    create the current spectral data file.

    Patterned after the provenance features of mzML that could also be mapped
    onto mzXML and other complete vendor readers.
    """

    @abc.abstractmethod
    def file_description(self):
        '''Describe the file and its components, as well
        as any content types it has.

        Returns
        -------
        :class:`~.FileInformation`
        '''
        return FileInformation()

    @abc.abstractmethod
    def instrument_configuration(self):
        '''Describe the different instrument components and configurations used
        to acquire scans in this run.

        Returns
        -------
        :class:`list` of :class:`~.InstrumentInformation`
        '''
        return []

    @abc.abstractmethod
    def data_processing(self):
        '''Describe any preprocessing steps applied to the data described by this
        instance.

        Returns
        -------
        :class:`list` of :class:`~.DataProcessingInformation`
        '''
        return []

    def software_list(self):
        '''Describe any software used on the data described by this instance.

        Returns
        -------
        :class:`list` of :class:`~.Software`
        '''
        return []



import unittest

from ms_deisotope.test.common import datafile
from ms_deisotope.data_source import infer_type
from ms_deisotope.data_source.metadata import file_information


class TestFileMetadata(unittest.TestCase):
    path = datafile("three_test_scans.mzML")

    @property
    def reader(self):
        return infer_type.MSFileLoader(self.path)

    def test_file_information(self):
        reader = self.reader
        finfo = reader.file_description()
        assert "MS1 spectrum" in finfo

    def test_source_file(self):
        id_fmt, fmt = file_information.SourceFile.guess_format(self.path)
        assert fmt == "mzML format"
        assert id_fmt == "no nativeID format"

        sf = file_information.SourceFile.from_path(self.path)
        assert not sf.has_checksum()
        sf.add_checksum('sha1')
        assert sf.has_checksum('sha1') and sf.has_checksum()
        assert sf.parameters["SHA-1"] == sf.checksum("sha1")
        assert sf.validate_checksum()

        other = sf.copy()
        assert sf == other


if __name__ == '__main__':
    unittest.main()



'''A collection of little command line utilities for inspecting mass
spectrum data.
'''
import os
import math

from collections import Counter

import click

import ms_deisotope

from ms_deisotope.feature_map import quick_index
from ms_deisotope.feature_map import scan_interval_tree

from ms_deisotope.clustering.scan_clustering import (
    iterative_clustering, ScanClusterWriter, ScanClusterReader, _DynamicallyLoadingResolver)

from ms_deisotope.qc.isolation import isolation_window_valid, is_isolation_window_empty

from ms_deisotope.data_source import (
    _compression, ScanProxyContext, MSFileLoader)
from ms_deisotope.data_source.scan import RandomAccessScanSource
from ms_deisotope.data_source.metadata.file_information import SourceFile

from ms_deisotope.output import ProcessedMzMLDeserializer

from ms_deisotope.tools import conversion, draw
from ms_deisotope.tools.utils import processes_option, is_debug_mode, register_debug_hook


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    '''A collection of utilities for inspecting and manipulating
    mass spectrometry data.
    '''
    pass


@cli.command("describe", short_help=("Produce a minimal textual description"
                                     " of a mass spectrometry data file"))
@click.argument('path', type=click.Path(exists=True))
@click.option("-d", "--diagnostics", is_flag=True,
              help="Run more diagnostics, greatly increasing runtime but producing additional informatoin")
def describe(path, diagnostics=False):
    '''Produces a minimal textual description of a mass spectrometry data file.
    '''
    click.echo("Describing \"%s\"" % (path,))
    try:
        sf = SourceFile.from_path(path)
    except IOError:
        click.echo("Could not open", err=True)
    if sf.file_format is None:
        click.echo("It doesn't appear to be a mass spectrometry data file")
        return -1
    click.echo("File Format: %s" % (sf.file_format, ))
    click.echo("ID Format: %s" % (sf.id_format, ))
    reader = MSFileLoader(path)
    if isinstance(reader, RandomAccessScanSource):
        click.echo("Format Supports Random Access: True")
        first_scan = reader[0]
        last_scan = reader[-1]
        click.echo("First Scan: %s at %0.3f minutes" % (first_scan.id, first_scan.scan_time))
        click.echo("Last Scan: %s at %0.3f minutes" % (last_scan.id, last_scan.scan_time))
    else:
        click.echo("Format Supports Random Access: False")
    try:
        finfo = reader.file_description()
        click.echo("Contents:")
        for key in finfo.contents:
            click.echo("    %s" % (key, ))
    except AttributeError:
        pass
    index_file_name = quick_index.ExtendedScanIndex.index_file_name(path)
    # Extra introspection if the extended index is available
    if os.path.exists(index_file_name):
        with open(index_file_name, 'rt') as fh:
            index = quick_index.ExtendedScanIndex.deserialize(fh)
        ms1_scans = len(index.ms1_ids)
        msn_scans = len(index.msn_ids)
        click.echo("MS1 Scans: %d" % (ms1_scans, ))
        click.echo("MSn Scans: %d" % (msn_scans, ))
        n_defaulted = 0
        n_orphan = 0

        charges = Counter()
        first_msn = float('inf')
        last_msn = 0
        for scan_info in index.msn_ids.values():
            n_defaulted += scan_info.get('defaulted', False)
            n_orphan += scan_info.get('orphan', False)
            charges[scan_info['charge']] += 1
            rt = scan_info['scan_time']
            if rt < first_msn:
                first_msn = rt
            if rt > last_msn:
                last_msn = rt
        click.echo("First MSn Scan: %0.3f minutes" % (first_msn,))
        click.echo("Last MSn Scan: %0.3f minutes" % (last_msn,))
        for charge, count in sorted(charges.items()):
            if not isinstance(charge, int):
                continue
            click.echo("Precursors with Charge State %d: %d" % (charge, count))
        if n_defaulted > 0:
            click.echo("Defaulted MSn Scans: %d" % (n_defaulted,))
        if n_orphan > 0:
            click.echo("Orphan MSn Scans: %d" % (n_orphan,))
    if diagnostics:
        reader.reset()
        scan_ids_with_invalid_isolation_window = []
        scan_ids_with_empty_isolation_window = []
        activation_counter = Counter()
        n_ms1 = 0
        n_msn = 0
        for precursor, products in reader:
            if precursor is not None:
                n_ms1 += 1
            for product in products:
                n_msn += 1
                if not isolation_window_valid(product):
                    scan_ids_with_invalid_isolation_window.append((precursor.id, product.id))
                if is_isolation_window_empty(product):
                    scan_ids_with_empty_isolation_window.append(
                        (precursor.id, product.id))
                activation_counter[product.activation] += 1

        click.echo("MS1 Spectra: %d" % (n_ms1, ))
        click.echo("MSn Spectra: %d" % (n_msn, ))
        click.echo("Invalid Isolation Windows: %d" % (
            len(scan_ids_with_invalid_isolation_window), ))
        click.echo("Empty Isolation Windows: %d" % (
            len(scan_ids_with_empty_isolation_window), ))

        click.echo("Activation Methods")
        for activation, count in activation_counter.items():
            if not activation.is_multiple_dissociation():
                click.echo("\t{method}:{energy} = {count}".format(
                    method=activation.method, energy=activation.energy, count=count))
            else:
                click.echo("\t{method}:{energy} = {count}".format(
                    method=','.join(map(str, activation.method)),
                    energy=','.join(map(str, activation.energies)), count=count))


@cli.command("byte-index", short_help='Build an external byte offset index for a mass spectrometry data file')
@click.argument('paths', type=click.Path(exists=True), nargs=-1)
def byte_index(paths):
    '''Build an external byte offset index for a mass spectrometry data file, saving time when
    opening the file with indexing enabled.

    Supported Formats: mzML, mzXML
    '''
    for path in paths:
        reader = ms_deisotope.MSFileLoader(path, use_index=False)
        try:
            fn = reader.prebuild_byte_offset_file
        except AttributeError:
            click.echo("\"%s\" does not support pre-indexing byte offsets" % (path,))
            return
        fn(path)


@cli.command("metadata-index", short_help='Build an external scan metadata index for a mass spectrometry data file')
@click.argument('paths', type=click.Path(exists=True), nargs=-1)
@processes_option
def metadata_index(paths, processes=4):
    '''Build an external scan metadata index for a mass spectrometry data file

    This extended index is saved in a separate JSON file that can be loaded with
    :class:`~.ExtendedScanIndex`. It includes the scan time of all scans, the precursor
    mass of MSn scans, as well as the relationships between precursor and product ion
    scans, as well as other details. See :class:`~.ExtendedScanIndex` for more information
    '''
    for path in paths:
        reader = MSFileLoader(path)
        try:
            fn = reader.prebuild_byte_offset_file
            if not reader.source._check_has_byte_offset_file():
                fn(path)
        except AttributeError:
            pass
        if processes > 1:
            progbar = click.progressbar(label='Building Index', length=100)
            acc = [0]

            def update_bar(x):
                '''Progress Bar update callback for :func:`~.quick_index.index`
                '''
                x = int(x * 100)
                x -= acc[0]  # pylint: disable=cell-var-from-loop
                progbar.update(x)  # pylint: disable=cell-var-from-loop
                acc[0] += x  # pylint: disable=cell-var-from-loop

            with progbar:
                update_bar(0.0)
                index, _ = quick_index.index(
                    reader, processes, progress_indicator=update_bar)
        else:
            index = quick_index.ExtendedScanIndex()
            reader.reset()
            for bunch in reader:
                index.add_scan_bunch(bunch)

        name = path
        index_file_name = index.index_file_name(name)
        with open(index_file_name, 'w') as fh:
            index.serialize(fh)


def _partial_ms_file_iterator(reader, start, end):
    for scan_bunch in reader.start_from_scan(index=start):
        if scan_bunch.precursor.index > end:
            break
        yield scan_bunch


class _MSMSIntervalTask(object):
    def __init__(self, time_radius, mz_lower, mz_higher):
        self.time_radius = time_radius
        self.mz_lower = mz_lower
        self.mz_higher = mz_higher

    def __call__(self, payload):
        reader, start, end = payload
        iterator = _partial_ms_file_iterator(reader, start, end)
        intervals = scan_interval_tree.extract_intervals(
            iterator, time_radius=self.time_radius,
            mz_lower=self.mz_lower, mz_higher=self.mz_higher)
        return intervals


@cli.command("msms-intervals", short_help=(
    'Build an interval tree over precursor isolation events in time and m/z space'))
@click.argument('paths', type=click.Path(exists=True), nargs=-1)
@click.option("-o", "--output", type=click.Path(writable=True, file_okay=True, dir_okay=False), required=False)
@processes_option
def msms_intervals(paths, processes=4, time_radius=5, mz_lower=2., mz_higher=3., output=None):
    '''Construct an interval tree spanning time and m/z domains where MSn spectra were acquired
    in the LC-MS map. The interval tree is serialized to JSON.
    '''
    interval_extraction = _MSMSIntervalTask(time_radius, mz_lower, mz_higher)
    interval_set = []
    total_work_items = len(paths) * processes * 4

    def _run():
        for path in paths:
            reader = MSFileLoader(path)
            chunk_out_of_order = quick_index.run_task_in_chunks(
                reader, processes, processes * 4, task=interval_extraction)
            for chunk in chunk_out_of_order:
                interval_set.extend(chunk)
                yield 0
    work_iterator = _run()
    with click.progressbar(work_iterator, length=total_work_items, label='Extracting Intervals') as g:
        for _ in g:
            pass
    tree = scan_interval_tree.ScanIntervalTree(scan_interval_tree.make_rt_tree(interval_set))
    if output is not None:
        with open(output, 'wt') as fh:
            tree.serialize(fh)
    else:
        stream = click.get_text_stream('stdout')
        tree.serialize(stream)
        stream.flush()


def _ensure_metadata_index(path):
    reader = ms_deisotope.MSFileLoader(path)
    name = path
    index_file_name = quick_index.ExtendedScanIndex.index_file_name(name)
    if not os.path.exists(index_file_name):
        click.secho("Building Index For %s" % (path, ), fg='yellow', err=True)
        index = quick_index.ExtendedScanIndex()
        reader.reset()
        for bunch in reader:
            index.add_scan_bunch(bunch)
        reader.reset()
        with open(index_file_name, 'w') as fh:
            index.serialize(fh)
    else:
        with open(index_file_name, 'rt') as fh:
            index = quick_index.ExtendedScanIndex.deserialize(fh)
    return reader, index


@cli.command("charge-states", short_help='Count the different precursor charge states in a mass spectrometry data file')
@click.argument("path", type=click.Path(exists=True))
def charge_states(path):
    '''Count the different precursor charge states in a mass spectrometry data file.

    This command will construct a metadata index if it is not found.
    '''
    _, index = _ensure_metadata_index(path)

    charges = Counter()
    for _, msn_info in index.msn_ids.items():
        charges[msn_info.charge] += 1
    for charge in sorted(charges, key=abs):
        click.echo("%d: %d" % (charge, charges[charge]))


def _binsearch(array, x):
    n = len(array)
    lo = 0
    hi = n

    while hi != lo:
        mid = (hi + lo) / 2
        y = array[mid][0]
        err = y - x
        if hi - lo == 1:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return


@cli.command("precursor-clustering", short_help='Cluster precursor masses in a mass spectrometry data file')
@click.argument("path", type=click.Path(exists=True))
def precursor_clustering(path, grouping_error=2e-5):
    '''Cluster precursor masses in a mass spectrometry data file.

    This command will construct a metadata index if it is not found.
    '''
    _, index = _ensure_metadata_index(path)
    points = []
    for _, msn_info in index.msn_ids.items():
        points.append((msn_info.neutral_mass, msn_info.intensity))
    points.sort(key=lambda x: x[1], reverse=1)
    centroids = []
    if not points:
        click.secho("No MS/MS detected", fg='yellow', err=True)
        return

    for point in points:
        if not centroids:
            centroids.append((point[0], [point]))
            continue
        i = _binsearch(centroids, point[0])
        centroid = centroids[i]
        err = (centroid[0] - point[0]) / point[0]
        if abs(err) < grouping_error:
            centroid[1].append(point)
        else:
            if err < 0:
                i += 1
            centroids.insert(i, (point[0], [point]))
    acc = 0
    nt = 0
    for centroid, obs in centroids:
        n = len(obs)
        if n == 1:
            continue
        mean = sum(p[0] for p in obs) / n
        var = sum([(p[0] - mean) ** 2 for p in obs]) / (n - 1)
        acc += (var * (n - 1))
        nt += (n - 1)
    for centroid, obs in centroids:
        click.echo("%f: %d" % (centroid, len(obs)))
    click.echo("MS/MS Precursor Mass Std. Dev.: %f Da" % (math.sqrt(acc / nt),))


@cli.command("spectrum-clustering", short_help=("Cluster MSn spectra in a mass spectrometry data file using"
                                                " cosine similarity"))
@click.argument('paths', type=click.Path(exists=True), nargs=-1)
@click.option("-m", "--precursor-error-tolerance", type=float, default=1e-5)
@click.option("-t", "--similarity-threshold", "similarity_thresholds", multiple=True, type=float)
@click.option("-o", "--output", "output_path", type=click.Path(writable=True, file_okay=True, dir_okay=False),
              required=False)
@click.option("-M", "--in-memory", is_flag=True, default=False, help=(
    "Whether to load the entire dataset into memory for better performance"))
@click.option("-D", "--deconvoluted", is_flag=True, default=False, help=(
    "Whether to assume the spectrum is deconvoluted or not"))
def spectrum_clustering(paths, precursor_error_tolerance=1e-5, similarity_thresholds=None, output_path=None,
                        in_memory=False, deconvoluted=False):
    '''Cluster spectra by precursor mass and cosine similarity.

    Spectrum clusters are written out to a text file recording
    cluster precursor mass, within-cluster similarity, and the
    source file and scan ID for each cluster member.
    '''
    if not similarity_thresholds:
        similarity_thresholds = [0.1, 0.4, 0.7]
    else:
        similarity_thresholds = sorted(similarity_thresholds)
    if output_path is None:
        output_path = "-"
    msn_scans = []
    n_spectra = 0

    with click.progressbar(paths, label="Indexing", item_show_func=lambda x: str(x) if x else '') as progbar:
        key_seqs = []
        for path in progbar:
            if deconvoluted:
                reader = ProcessedMzMLDeserializer(path)
                index = reader.extended_index
            else:
                reader, index = _ensure_metadata_index(path)
            key_seqs.append((reader, index))
            n_spectra += len(index.msn_ids)

    with click.progressbar(label="Loading Spectra", length=n_spectra,
                           item_show_func=lambda x: str(x) if x else '') as progbar:
        for reader, index in key_seqs:
            if not in_memory:
                proxy_context = ScanProxyContext(reader)
                pinfo_map = {
                    pinfo.product_scan_id: pinfo for pinfo in
                    index.get_precursor_information()
                }
                for i in index.msn_ids:
                    progbar.current_item = i
                    progbar.update(1)
                    scan = proxy_context(i)
                    scan.precursor_information = pinfo_map[i]
                    msn_scans.append(scan)
            else:
                for i in index.msn_ids:
                    progbar.current_item = i
                    progbar.update(1)
                    scan = reader.get_scan_by_id(i)
                    if scan.peak_set is None and not deconvoluted:
                        scan.pick_peaks()
                    msn_scans.append(scan)

    click.echo("Begin Clustering", err=True)
    clusters = iterative_clustering(
        msn_scans, precursor_error_tolerance, similarity_thresholds)
    click.echo("Clusering Finished", err=True)
    by_size = Counter()
    for cluster in clusters:
        by_size[len(cluster)] += 1
    click.echo("Clusters: {:d}".format(len(clusters)), err=True)
    for key, value in sorted(by_size.items()):
        click.echo("Size {:d}: {:d}".format(key, value), err=True)
    with click.open_file(output_path, mode='w') as outfh:
        writer = ScanClusterWriter(outfh)
        for cluster in clusters:
            writer.save(cluster)


@cli.command('cluster-statistics')
@click.argument('path', type=click.Path(readable=True, dir_okay=False))
def cluster_evaluation(path):
    """Calculate some statistics describing the clustered mass spectra described in
    the specified cluster file.
    """
    with click.open_file(path) as fh:
        dirname = os.path.dirname(path)
        reader = ScanClusterReader(fh, _DynamicallyLoadingResolver(dirname))
        clusters = list(reader)
        by_size = Counter()
        for cluster in clusters:
            by_size[len(cluster)] += 1
        click.echo("Clusters: {:d}".format(len(clusters)))
        for key, value in sorted(by_size.items()):
            click.echo("Size {:d}: {:d}".format(key, value))


if _compression.has_idzip:
    @cli.command("idzip", short_help='Compress a file with idzip, a gzip-compatible format with random access support')
    @click.argument('path', type=click.Path(allow_dash=True, readable=True))
    @click.option("-o", "--output", type=click.Path(writable=True, file_okay=True, dir_okay=False),
                  required=False, help="The output stream to write to. Defaults to STDOUT if omitted.")
    def idzip_compression(path, output):
        '''Compress a file using  idzip, a gzip-compatible format with random access support.
        '''
        if output is None:
            output = '-'
        with click.open_file(output, mode='wb') as outfh:
            writer = _compression.GzipFile(fileobj=outfh, mode='wb')
            with click.open_file(path, 'rb') as infh:
                buffer_size = _compression.WRITE_BUFFER_SIZE
                chunk = infh.read(buffer_size)
                while chunk:
                    writer.write(chunk)
                    chunk = infh.read(buffer_size)
            writer.close()



def _mount_group(group):
    try:
        for name, command in group.commands.items():
            cli.add_command(command, name)
    except Exception as e:
        click.secho("%r occurred while loading additional tools from %r" % (e, group), err=True, fg='yellow')


_mount_group(conversion.ms_conversion)

cli.add_command(draw.draw)

main = cli.main

if is_debug_mode():
    register_debug_hook()

if __name__ == '__main__':
    click.secho("Running Debug Mode", fg='yellow')
    register_debug_hook()
    main()



import unittest
import logging
import numpy as np

import brainpy

from ms_deisotope.data_source import common, mzml, MSFileLoader
from ms_deisotope.averagine import peptide, glycopeptide, TheoreticalIsotopicPattern
from ms_peak_picker import reprofile
from ms_deisotope.deconvolution import (
    deconvolute_peaks, AveragineDeconvoluter,
    AveraginePeakDependenceGraphDeconvoluter,
    CompositionListDeconvoluter,
    CompositionListPeakDependenceGraphDeconvoluter,
    count_placeholders, drop_placeholders,
    ChargeIterator)
from ms_deisotope.scoring import PenalizedMSDeconVFitter, MSDeconVFitter
from brainpy import neutral_mass

from ms_deisotope.test.test_scan import (
    make_profile, points, fwhm,
    FittedPeak)
from ms_deisotope.test.common import datafile


logger = logging.getLogger('ms_deisotope.test.test_deconvolution')


class TestAveragineDeconvolution(unittest.TestCase):
    def make_scan(self):
        mz, intensity = make_profile(points, fwhm)
        scan = common.Scan(
            {
                "id": "test-scan",
                "index": 0,
                "m/z array": mz,
                "ms level": 1,
                "scan time": 0.0,
                "intensity array": intensity,
                "profile spectrum": "",
                "positive scan": "",
            },
            mzml.MzMLDataInterface())
        scan.pick_peaks()
        for point in points:
            self.assertIsNotNone(scan.peak_set.has_peak(point[0]))
        return scan

    def test_quick_charge(self):
        scan = self.make_scan()
        peaks = scan.peak_set
        peak = peaks[0]
        charge_states = ChargeIterator(1, 8)
        charge_states.sequence_from_quickcharge(peaks, peak)
        states = list(charge_states)
        self.assertEqual(states, [1, 3])

    def test_deconvolution(self):
        scan = self.make_scan()
        algorithm_type = AveragineDeconvoluter
        deconresult = deconvolute_peaks(
            scan.peak_set, {
                "averagine": peptide,
                "scorer": PenalizedMSDeconVFitter(5., 1.),
                "use_subtraction": False
            }, deconvoluter_type=algorithm_type)
        dpeaks = deconresult.peak_set
        assert len(dpeaks) == 6
        for point in points:
            peak = dpeaks.has_peak(neutral_mass(point[0], point[1]))
            self.assertIsNotNone(peak)

    def test_graph_deconvolution(self):
        scan = self.make_scan()
        scan.pick_peaks()
        self.assertIsNotNone(scan.peak_set)
        algorithm_type = AveraginePeakDependenceGraphDeconvoluter
        deconresult = deconvolute_peaks(
            scan.peak_set, {
                "averagine": peptide,
                "scorer": PenalizedMSDeconVFitter(5., 1.)
            }, deconvoluter_type=algorithm_type)
        dpeaks = deconresult.peak_set
        assert len(dpeaks) == 2
        deconvoluter = deconresult.deconvoluter
        for point in points:
            peak = dpeaks.has_peak(neutral_mass(point[0], point[1]))
            self.assertIsNotNone(peak)
            fp = scan.has_peak(peak.mz)
            self.assertAlmostEqual(
                deconvoluter.peak_dependency_network.find_solution_for(fp).mz,
                peak.mz, 3)


class TestCompositionListDeconvolution(unittest.TestCase):
    compositions = [
        brainpy.parse_formula('C84H138N6O62'),
        brainpy.parse_formula('C90H148N6O66'),
        brainpy.parse_formula('C90H148N6O69S1'),
        brainpy.parse_formula('C62H94N4O69S6'),
    ]

    charges = [
        ((2, 100), (3, 250),),
        ((2, 75), (3, 200),),
        ((2, 90), (3, 270),),
        ((3, 20), (4, 90), (5, 45))
    ]

    def make_tids(self):
        tids = []
        ions = []
        for comp, charges in zip(self.compositions, self.charges):
            for charge, abundance in charges:
                tid = brainpy.isotopic_variants(comp, charge=-charge)
                tid = TheoreticalIsotopicPattern(tid, tid[0].mz)
                tid.scale_raw(abundance * 100)
                tids.append(tid)
                ions.append((comp, -charge))
        return tids, ions

    @staticmethod
    def get_nearest_index(query_mz, tid_list):
        best_index = None
        best_error = float('inf')

        for i, tid in enumerate(tid_list):
            error = abs(tid.monoisotopic_mz - query_mz)
            if error < best_error:
                best_error = error
                best_index = i
        return best_index

    def make_scan(self):
        peaks = []
        tids, ions = self.make_tids()
        list(map(peaks.extend, tids))
        peaks.sort(key=lambda x: x.mz)

        mz = np.array([0])
        intensity = np.array([0])

        fpeaks = []
        for p in peaks:
            fpeaks.append(FittedPeak(
                mz=p.mz, intensity=p.intensity, signal_to_noise=p.intensity, peak_count=-1, index=-1,
                full_width_at_half_max=fwhm, area=p.intensity))
        mz, intensity = reprofile(fpeaks)
        scan = common.Scan(
            {
                "id": "test-scan",
                "index": 0,
                "m/z array": mz,
                "intensity array": intensity,
                "ms level": 1,
                "scan time": 0.0,
                "profile spectrum": "",
                "negative scan": "",
            },
            mzml.MzMLDataInterface())
        return scan

    def test_deconvolution(self):
        scan = self.make_scan()
        scan.pick_peaks()
        self.assertIsNotNone(scan.peak_set)
        algorithm_type = CompositionListDeconvoluter
        decon_config = {
            "composition_list": self.compositions,
            "scorer": PenalizedMSDeconVFitter(5., 2.),
            # "scorer": MSDeconVFitter(5.),
            "use_subtraction": False
        }
        deconresult = deconvolute_peaks(
            scan.peak_set, decon_config, charge_range=(-1, -8), deconvoluter_type=algorithm_type)
        dpeaks = deconresult.peak_set
        n_cases = sum(map(len, self.charges))
        if not (len(dpeaks) == n_cases):
            tids, ions = self.make_tids()
            tids, ions = zip(*sorted(zip(tids, ions), key=lambda x: x[0].monoisotopic_mz))
            seen = set()
            for i, dp in enumerate(sorted(dpeaks, key=lambda x: x.mz)):
                ix = self.get_nearest_index(dp.mz, tids)
                logger.warning("%0.3f %d %0.3f %r (Matched %d)", dp.neutral_mass, dp.charge, dp.score, dp.solution, ix)
                seen.add(ix)
            indices = set(range(len(ions)))
            missed = list(indices - seen)
            deconvoluter = algorithm_type(scan.peak_set.clone(), **decon_config)
            for ix in missed:
                tid = deconvoluter.generate_theoretical_isotopic_cluster(*ions[ix])
                assert np.isclose(sum(p.intensity for p in tid), 1.0)
                monoisotopic_peak = deconvoluter.peaklist.has_peak(tid[0].mz, 2e-5)
                if monoisotopic_peak is not None:
                    tid = deconvoluter.recalibrate_theoretical_mz(tid, monoisotopic_peak.mz)
                eid = deconvoluter.match_theoretical_isotopic_distribution(
                    tid.peaklist, 2e-5)
                missed_peaks = count_placeholders(eid)
                deconvoluter.scale_theoretical_distribution(tid, eid)
                score = deconvoluter.scorer.evaluate(deconvoluter.peaklist, eid, tid.peaklist)
                fit_record = deconvoluter.fit_composition_at_charge(*ions[ix])
                eid = fit_record.experimental
                tid = fit_record.theoretical
                rep_eid = drop_placeholders(eid)
                validation = (len(rep_eid) < 2), (len(rep_eid) < len(tid) / 2.), (
                    len(rep_eid) == 1 and fit_record.charge > 1)
                composition, charge = ions[ix]
                logger.warning("Missed %r %d (%d missed peaks, score = %0.3f, record = %r, validation = %r)" % (
                    composition, charge, missed_peaks, score, fit_record, validation))
            assert not missed

    def test_graph_deconvolution(self):
        scan = self.make_scan()
        scan.pick_peaks()
        self.assertIsNotNone(scan.peak_set)
        algorithm_type = CompositionListPeakDependenceGraphDeconvoluter
        decon_config = {
            "composition_list": self.compositions,
            "scorer": PenalizedMSDeconVFitter(5., 2.),
            "use_subtraction": True
        }
        deconresult = deconvolute_peaks(
            scan.peak_set, decon_config, charge_range=(-1, -8), deconvoluter_type=algorithm_type)
        dpeaks = deconresult.peak_set
        n_cases = sum(map(len, self.charges))
        # assert len(dpeaks) == n_cases
        if not (len(dpeaks) == n_cases):
            tids, ions = self.make_tids()
            tids, ions = zip(*sorted(zip(tids, ions), key=lambda x: x[0].monoisotopic_mz))
            seen = set()
            for i, dp in enumerate(sorted(dpeaks, key=lambda x: x.mz)):
                ix = self.get_nearest_index(dp.mz, tids)
                logger.warning("%0.3f %d %0.3f %r (Matched %d)", dp.neutral_mass, dp.charge, dp.score, dp.solution, ix)
                seen.add(ix)
            indices = set(range(len(ions)))
            missed = list(indices - seen)
            deconvoluter = algorithm_type(scan.peak_set.clone(), **decon_config)
            for ix in missed:
                tid = deconvoluter.generate_theoretical_isotopic_cluster(*ions[ix])
                assert np.isclose(sum(p.intensity for p in tid), 1.0)
                monoisotopic_peak = deconvoluter.peaklist.has_peak(tid[0].mz, 2e-5)
                if monoisotopic_peak is not None:
                    tid = deconvoluter.recalibrate_theoretical_mz(tid, monoisotopic_peak.mz)
                eid = deconvoluter.match_theoretical_isotopic_distribution(
                    tid.peaklist, 2e-5)
                missed_peaks = count_placeholders(eid)
                deconvoluter.scale_theoretical_distribution(tid, eid)
                score = deconvoluter.scorer.evaluate(deconvoluter.peaklist, eid, tid.peaklist)
                fit_record = deconvoluter.fit_composition_at_charge(*ions[ix])
                eid = fit_record.experimental
                tid = fit_record.theoretical
                rep_eid = drop_placeholders(eid)
                validation = (len(rep_eid) < 2), (len(rep_eid) < len(tid) / 2.), (
                    len(rep_eid) == 1 and fit_record.charge > 1)
                composition, charge = ions[ix]
                logger.warning("Missed %r %d (%d missed peaks, score = %0.3f, record = %r, validation = %r)" % (
                    composition, charge, missed_peaks, score, fit_record, validation))
            assert not missed


class TestSolutionRetrieval(unittest.TestCase):
    def make_scan(self):
        complex_compressed_mzml = datafile("20150710_3um_AGP_001_29_30.mzML.gz")
        reader = MSFileLoader(complex_compressed_mzml)
        bunch = next(reader)
        return bunch

    def target_envelopes(self):
        envelopes = [
            [1161.00927734375,
             1161.25830078125,
             1161.508544921875,
             1161.7586669921875,
             1162.0111083984375,
             1162.2607421875,
             1162.5128173828125],
            [1197.52197265625,
             1197.7735595703125,
             1198.024169921875,
             1198.2742919921875,
             1198.524658203125,
             1198.774169921875,
             1199.02490234375],
            [929.0088500976562,
             929.2091674804688,
             929.4107055664062,
             929.6107788085938,
             929.8109741210938,
             930.01171875,
             930.213134765625],
            [1088.234619140625,
             1088.4853515625,
             1088.7354736328125,
             1088.987060546875,
             1089.237060546875,
             1089.4873046875,
             1089.7379150390625],
            [1002.0338745117188,
             1002.2335815429688,
             1002.4354858398438,
             1002.63671875,
             1002.837646484375,
             1003.0364990234375,
             1003.2337646484375],
        ]
        return envelopes

    def test_retrieve_deconvolution_solution(self):
        bunch = self.make_scan()
        scan = bunch.precursor
        scan.pick_peaks()
        ms1_deconvolution_args = {
            "averagine": glycopeptide,
            "scorer": PenalizedMSDeconVFitter(20., 2.),
        }
        priorities = []
        for product in bunch.products:
            priorities.append(scan.has_peak(product.precursor_information.mz))
        algorithm_type = AveraginePeakDependenceGraphDeconvoluter
        deconresult = deconvolute_peaks(
            scan.peak_set, ms1_deconvolution_args,
            priority_list=priorities,
            deconvoluter_type=algorithm_type)
        dpeaks = deconresult.peak_set
        deconvoluter = deconresult.deconvoluter
        priority_results = deconresult.priorities
        reference_deconvoluter = algorithm_type(scan.peak_set.clone(), **ms1_deconvolution_args)
        for i, result in enumerate(priority_results):
            query = priorities[i].mz
            if result is None:
                logger.warn("Query %d (%f) had no result", i, query)
                raw_peaks = scan.peak_set.between(query - 2, query + 3)
                anchor_peak = scan.peak_set.has_peak(query)
                deconvoluted_peaks = dpeaks.between(query - 2, query + 3, use_mz=True)
                possible_solutions = reference_deconvoluter._fit_all_charge_states(anchor_peak)
                sols = []
                logger.warn("Possible Solutions %r", possible_solutions)
                if not possible_solutions:
                    for charge in [3, 4, 5]:
                        tid = reference_deconvoluter.averagine.isotopic_cluster(
                            anchor_peak.mz, charge)
                        assert np.isclose(tid.monoisotopic_mz, anchor_peak.mz)
                        assert np.isclose(sum(p.intensity for p in tid), 1.0)
                        eid = reference_deconvoluter.match_theoretical_isotopic_distribution(
                            tid.peaklist, error_tolerance=2e-5)
                        assert len(eid) == len(tid)
                        record = reference_deconvoluter._evaluate_theoretical_distribution(
                            eid, tid, anchor_peak, charge)
                        sols.append(record)
                    logger.warn("Manually Generated Solutions %r", sols)
                assert anchor_peak is not None and raw_peaks and (possible_solutions or sols) and not deconvoluted_peaks
                assert deconvoluter.peak_dependency_network.find_solution_for(anchor_peak) is not None
                assert dpeaks.has_peak(query, use_mz=True)
                # error out
                assert result is not None
            else:
                assert 0 <= abs(result.mz - query) < 1
                anchor_peak = scan.peak_set.has_peak(query)
                assert deconvoluter.peak_dependency_network.find_solution_for(anchor_peak) is not None


if __name__ == '__main__':
    unittest.main()



# pragma: no cover
from ms_deisotope.utils import Base
from ms_deisotope.data_source.common import ScanBunch, ScanIterator


class ScanSerializerBase(object):
    """A common base class for all types which serialize :class:`~.Scan`-like objects to
    disk.

    These types support the context manager protocol, controlling their closing behavior.
    """

    def __init__(self, *args, **kwargs):
        super(ScanSerializerBase, self).__init__(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def save_scan_bunch(self, bunch, **kwargs):
        """Save a single :class:`~.ScanBunch` to the
        writing stream, recording any information relating
        the precursors and products to each other.

        May call :meth:`save_scan` multiple times.

        Parameters
        ----------
        bunch : :class:`~.ScanBunch`
            The scan bunch to save
        """
        self.save_scan(bunch.precursor, **kwargs)
        for prod in bunch.products:
            self.save_scan(prod, **kwargs)

    def save_scan(self, scan, **kwargs):
        """Save a single scan to the writing stream.

        Parameters
        ----------
        scan : :class:`ScanBase`
            The scan to save
        """
        raise NotImplementedError()

    def save(self, bunch, **kwargs):
        """Save any scan information in `bunch`.

        This method can handle :class:`~.ScanBunch` or :class:`ScanBase`
        instances, dispatching to the appropriate logic.

        Parameters
        ----------
        bunch : :class:`~.ScanBunch` or :class:`ScanBase`
            The scan data to save. May be a collection of related scans or a single scan.

        See Also
        --------
        :meth:`save_scan`
        :meth:`save_scan_bunch`
        """
        if isinstance(bunch, ScanBunch):
            self.save_scan_bunch(bunch, **kwargs)
        else:
            self.save_scan(bunch, **kwargs)

    def close(self):
        """Finish writing scan data, write any pending metadata
        and close the file stream.

        May call :meth:`complete`.
        """
        self.complete()

    def complete(self):
        """Perform any final operations after all spectra have been
        written, adding any trailing content necessary for the format.

        May be called from :meth:`close`, but before the writing stream
        has actually been closed.
        """
        pass


class ScanDeserializerBase(object):
    def __init__(self, *args, **kwargs):
        super(ScanDeserializerBase, self).__init__(*args, **kwargs)

    def __next__(self):
        return self.next()

    def next(self):
        raise NotImplementedError()

    def get_scan_by_id(self, scan_id):
        raise NotImplementedError()

    def get_scan_by_time(self, rt, require_ms1=False):
        raise NotImplementedError()

    def get_scan_by_index(self, index):
        raise NotImplementedError()


ScanIterator.register(ScanDeserializerBase)


class SampleRun(Base):

    def __init__(self, name, uuid, completed=True, sample_type=None, **kwargs):
        self.name = name
        self.uuid = uuid
        self.sample_type = sample_type
        self.completed = completed
        self.parameters = kwargs



import re
from collections import defaultdict

import numpy as np

from ms_deisotope.utils import Base

from .metadata.instrument_components import (
    component, ComponentGroup, InstrumentInformation)
from .metadata.file_information import (
    FileInformation, MS_MS1_Spectrum, MS_MSn_Spectrum)

from .common import ScanFileMetadataBase

analyzer_pat = re.compile(r"(?P<mass_analyzer_type>ITMS|TQMS|SQMS|TOFMS|FTMS|SECTOR)")
polarity_pat = re.compile(r"(?P<polarity>[\+\-])")
point_type_pat = re.compile(r"(?P<point_type>[CP])")
ionization_pat = re.compile(r"(?P<ionization_type>EI|CI|FAB|APCI|ESI|APCI|NSI|TSP|FD|MALDI|GD)")
scan_type_pat = re.compile(r"(?P<scan_type>FULL|SIM|SRM|CRM|Z|Q1MS|Q3MS)")
ms_level_pat = re.compile(r" ms(?P<level>\d*) ")
activation_pat = re.compile(
    r"""(?:(?P<isolation_mz>\d+\.\d*)@
        (?P<activation_type>[a-z]+)
        (?P<activation_energy>\d*\.?\d*))""", re.VERBOSE)
activation_mode_pat = re.compile(
    r"""(?P<activation_type>[a-z]+)
        (?P<activation_energy>\d*\.\d*)""", re.VERBOSE)
scan_window_pat = re.compile(
    r"""
    \[(?P<scan_start>[0-9\.]+)-(?P<scan_end>[0-9\.]+)\]
    """, re.VERBOSE)

analyzer_map = {
    'FTMS': component("orbitrap"),
    "ITMS": component("ion trap"),
    "SQMS": component("quadrupole"),
    "TQMS": component("quadrupole"),
    "TOFMS": component("time-of-flight"),
    "SECTOR": component("magnetic sector")
}


ionization_map = {
    "EI": component("electron ionization"),
    "CI": component("chemical ionization"),
    "FAB": component("fast atom bombardment ionization"),
    "ESI": component("electrospray ionization"),
    "NSI": component("nanoelectrospray"),
    "APCI": component("atmospheric pressure chemical ionization"),
    "TSP": component("thermospray ionization"),
    "FD": component("field desorption"),
    "MALDI": component("matrix assisted laser desorption ionization"),
    "GD": component("glow discharge ionization"),
}


inlet_map = {
    "FAB": component("continuous flow fast atom bombardment"),
    "ESI": component("electrospray inlet"),
    "NSI": component("nanospray inlet"),
    "TSP": component("thermospray inlet"),
}


class FilterString(str):
    def __init__(self, value):
        self.data = self._parse()

    def get(self, key):
        return self.data.get(key)

    def _parse(self):
        return filter_string_parser(self)


def filter_string_parser(line):
    """Parses instrument information from Thermo's filter string

    Parameters
    ----------
    line : str
        The filter string associated with a scan

    Returns
    -------
    dict
        Fields extracted from the filter string
    """
    words = line.upper().split(" ")
    values = dict()
    i = 0
    values['supplemental_activation'] = " sa " in line
    ms_level_info = ms_level_pat.search(line)
    if ms_level_info is not None:
        ms_level_data = ms_level_info.groupdict()
        level = ms_level_data.get("level")
        if level != "":
            parts = line[ms_level_info.end():].split(" ")
            tandem_sequence = []
            for part in parts:
                activation_info = activation_pat.search(part)
                if activation_info is not None:
                    activation_info = activation_info.groupdict()
                    activation_event = dict()
                    activation_event["isolation_mz"] = float(activation_info['isolation_mz'])
                    activation_event["activation_type"] = [activation_info['activation_type']]
                    activation_event["activation_energy"] = [float(activation_info['activation_energy'])]
                    if part.count("@") > 1:
                        act_events = activation_mode_pat.finditer(part)
                        # discard the first match which we already recorded
                        next(act_events)
                        for match in act_events:
                            act_type, act_energy = match.groups()
                            act_energy = float(act_energy)
                            activation_event["activation_type"].append(act_type)
                            activation_event['activation_energy'].append(act_energy)
                    tandem_sequence.append(activation_event)
            values['ms_level'] = int(level)
            values['tandem_sequence'] = tandem_sequence

    scan_window_info = scan_window_pat.search(line)
    if scan_window_info is not None:
        values['scan_window'] = (float(scan_window_info.group(1)), float(scan_window_info.group(2)))

    try:
        word = words[i]
        i += 1
        analyzer_info = analyzer_pat.search(word)
        if analyzer_info is not None:
            values['analyzer'] = analyzer_info.group(0)
            word = words[i]
            i += 1
        polarity_info = polarity_pat.search(word)
        if polarity_info is not None:
            polarity_sigil = polarity_info.group(0)
            if polarity_sigil == "+":
                polarity = 1
            elif polarity_sigil == "-":
                polarity = -1
            else:
                polarity = 0
            values["polarity"] = polarity
            word = words[i]
            i += 1
        if word in "PC":
            if word == 'P':
                values['peak_mode'] = 'profile'
            else:
                values['peak_mode'] = 'centroid'
            word = words[i]
            i += 1
        ionization_info = ionization_pat.search(word)
        if ionization_info is not None:
            values['ionization'] = ionization_info.group(0)
            word = words[i]
            i += 1

        return values
    except IndexError:
        return values


_id_template = "controllerType=0 controllerNumber=1 scan="


def _make_id(scan_number):
    try:
        return "%s%d" % (_id_template, (scan_number))
    except TypeError:
        return None


def _parse_id(scan_id):
    return int(scan_id.replace(_id_template, ""))


class _RawFileMetadataLoader(ScanFileMetadataBase):
    def _get_instrument_serial_number(self):
        return ''

    def _get_instrument_model_name(self):
        return ''

    def _build_scan_type_index(self):
        self.make_iterator(grouped=False)
        index = defaultdict(int)
        analyzer_counter = 1
        analyzer_confs = dict()
        for scan in self: # pylint: disable=not-an-iterable
            index[scan.ms_level] += 1
            fline = self._filter_string(scan._data)
            analyzer = analyzer_map[fline.data['analyzer']]
            try:
                analyzer_confs[analyzer]
            except KeyError:
                analyzer_confs[analyzer] = analyzer_counter
                analyzer_counter += 1
        self.reset()
        self._scan_type_index = index
        self._analyzer_to_configuration_index = analyzer_confs

    def _get_instrument_info(self):
        scan = self.get_scan_by_index(0)
        filter_string = self._filter_string(scan._data)
        ionization_label = filter_string.data.get("ionization")
        try:
            ionization = ionization_map[ionization_label]
        except KeyError:
            ionization = ionization_map['ESI']
        try:
            inlet = inlet_map[ionization_label]
        except KeyError:
            inlet = None

        source_group = ComponentGroup("source", [], 1)
        source_group.add(ionization)
        if inlet is not None:
            source_group.add(inlet)
        configs = []
        for analyzer, counter in sorted(self._analyzer_to_configuration_index.items(), key=lambda x: x[1]):
            analyzer_group = ComponentGroup('analyzer', [analyzer], 2)
            # this is the most common Thermo detector, but it is not universal. To get this right,
            # we'd need to reproduce the Proteowizard conversion table @
            # Thermo::(Reader_Thermo_Detail::)createInstrumentConfigurations
            detector_group = ComponentGroup("detector", [component('inductive detector')], 3)
            configs.append(InstrumentInformation(
                counter, [source_group, analyzer_group, detector_group],
                self._get_instrument_model_name(), self._get_instrument_serial_number())
            )
        self._instrument_config = {
            c.id: c for c in configs
        }
        return configs

    def instrument_configuration(self):
        return sorted(self._instrument_config.values(), key=lambda x: x.id)

    def file_description(self):
        fi = FileInformation({}, [])
        fi.add_file(self.source_file)
        sf = fi.source_files[0]
        sf.add_checksum("sha1")
        if 1 in self._scan_type_index:
            fi.add_content(MS_MS1_Spectrum)
        scan_types = sorted(self._scan_type_index, reverse=True)
        if scan_types:
            if scan_types[0] > 1:
                fi.add_content(MS_MSn_Spectrum)
        return fi

    def data_processing(self):
        return []


class _InstrumentMethod(object):
    def __init__(self, method_text):
        self.text = method_text
        (self.isolation_width_by_segment_and_event,
         self.isolation_width_by_segment_and_ms_level) = method_parser(self.text)

    def isolation_width_for(self, segment, event=None, ms_level=None):
        if event is not None:
            try:
                width = self.isolation_width_by_segment_and_event[segment][event]
                return width
            except KeyError:
                return 0.0
        elif ms_level is not None:
            try:
                width = self.isolation_width_by_segment_and_ms_level[segment][ms_level]
                return width
            except KeyError:
                return 0.0
        else:
            raise ValueError("One of event or ms_level must not be None!")


def method_parser(method_text):
    scan_segment_re = re.compile(r"\s*Segment (\d+) Information\s*")
    scan_event_re = re.compile(r"\s*(\d+):.*")
    scan_event_isolation_width_re = re.compile(r"\s*Isolation Width:\s*(\S+)\s*")
    scan_event_iso_w_re = re.compile(r"\s*MS.*:.*\s+IsoW\s+(\S+)\s*")
    repeated_event_re = re.compile(r"\s*Scan Event (\d+) repeated for top (\d+)\s*")
    default_isolation_width_re = re.compile(r"\s*MS(\d+) Isolation Width:\s*(\S+)\s*")

    scan_segment = 1
    scan_event = 0
    scan_event_details = False
    data_dependent_settings = False

    isolation_width_by_segment_and_event = defaultdict(dict)
    isolation_width_by_segment_and_ms_level = defaultdict(dict)

    for line in method_text.splitlines():
        match = scan_segment_re.match(line)

        if match:
            scan_segment = int(match.group(1))
            continue

        if "Scan Event Details" in line:
            scan_event_details = True
            continue

        if scan_event_details:
            match = scan_event_re.match(line)
            if match:
                scan_event = int(match.group(1))
                continue

            match = scan_event_isolation_width_re.match(line)
            if match:
                isolation_width_by_segment_and_event[scan_segment][scan_event] = float(match.group(1))
                continue

            match = scan_event_iso_w_re.match(line)
            if match:
                isolation_width_by_segment_and_event[scan_segment][scan_event] = float(match.group(1))
                continue

            match = repeated_event_re.match(line)
            if match:
                repeated_event = int(match.group(1))
                repeat_count = int(match.group(2))
                repeated_width = isolation_width_by_segment_and_event[scan_segment][repeated_event]
                for i in np.arange(repeated_width + 1, repeat_count + repeated_width):
                    isolation_width_by_segment_and_event[scan_segment][i] = repeated_width
                continue

            if not line.strip():
                scan_event_details = False

        if "Data Dependent Settings" in line:
            data_dependent_settings = True
            continue

        if data_dependent_settings:
            match = default_isolation_width_re.match(line)
            if match:
                ms_level = int(match.group(1))
                width = float(match.group(2))
                isolation_width_by_segment_and_ms_level[scan_segment][ms_level] = width
                continue

            if not line.strip():
                data_dependent_settings = False

    return isolation_width_by_segment_and_event, isolation_width_by_segment_and_ms_level


class ThermoRawScanPtr(Base):
    def __init__(self, scan_number):
        self.scan_number = scan_number
        self.filter_string = None

    def validate(self, source):
        try:
            source._scan_time(self)
            return True
        except IOError:
            return False



'''A common set of methods that are shared by
all :mod:`pyteomics`-based XML file readers.
'''

import warnings

from lxml import etree
from lxml.etree import XMLSyntaxError

from pyteomics import xml
from pyteomics.xml import unitfloat

from .common import (
    RandomAccessScanSource)
from ._compression import get_opener


def in_minutes(x):
    '''Convert a time quantity to minutes

    Parameters
    ----------
    x: unitfloat
        A float representing a quantity of time annotated with a time unit

    Returns
    -------
    unitfloat:
        The time after conversion to minutes
    '''
    try:
        unit = x.unit_info
    except AttributeError:
        return x
    if unit == 'minute':
        return x
    elif unit == 'second':
        y = unitfloat(x / 60., 'minute')
        return y
    elif unit == 'hour':
        y = unitfloat(x * 60, 'minute')
        return y
    else:
        warnings.warn("Time unit %r not recognized" % unit)
    return x


class XMLReaderBase(RandomAccessScanSource):
    '''A common implementation of :mod:`pyteomics`-based XML file formats.

    Attributes
    ----------
    index: :class:`pyteomics.xml.ByteEncodingOrderedDict`
        The byte offset index used to achieve fast random access
    '''

    _parser_cls = None


    @classmethod
    def prebuild_byte_offset_file(cls, path):
        """Parse the file given by `path`, generating a byte offset index in
        JSON format and save it to disk for future use.

        This method is intended to provide a way to save time during repeated
        instantiation of this type over the same file by removing the need to
        do a full scan of the file to rebuild of the offset index each time.

        .. note::

            This assumes that `path` is either a path to a file in a directory
            which the invoking user has read and write access to, or that it is
            a file-like object whose `name` attribute gives a path that satisfies
            the same requirements.

        Parameters
        ----------
        path : :class:`str` or file-like
            The path to the file to index, or a file-like object with a name attribute.
        """
        return cls._parser_cls.prebuild_byte_offset_file(get_opener(path))

    @property
    def index(self):
        '''The byte offset index used to achieve fast random access.

        Maps :class:`~.ScanBase` IDs to the byte offsets, implying
        the order the scans reside in the file.

        Returns
        -------
        :class:`pyteomics.xml.ByteEncodingOrderedDict`
        '''
        return self._source._offset_index

    @property
    def source(self):
        '''The file parser that this reader consumes.
        '''
        return self._source

    @source.setter
    def source(self, value):
        self._source = value

    def close(self):
        '''Close the underlying reader.
        '''
        if self.source is not None:
            self.source.close()
            self.source = None

    def __del__(self):
        try:
            self.close()
        except AttributeError:
            pass

    def __len__(self):
        return len(self.index)

    def reset(self):
        """Reset the object, clearing out any existing
        state.

        This resets the underlying file iterator, then
        calls :meth:`make_iterator`, and clears the scan
        cache.
        """
        self._source.reset()
        try:
            self.source.seek(0)
        except (IOError, AttributeError):
            pass
        self.make_iterator(None)
        self.initialize_scan_cache()

    def _validate(self, scan):
        raise NotImplementedError()

    def _make_default_iterator(self):
        return iter(self._source)

    def next(self):
        try:
            return next(self._producer)
        except XMLSyntaxError:
            raise StopIteration(
                "This iterator may need to be reset by calling `reset` to continue using it after"
                " using a random-access function like `get_by_id`")

    def __next__(self):
        return self.next()

    def get_scan_by_id(self, scan_id):
        """Retrieve the scan object for the specified scan id.

        If the scan object is still bound and in memory somewhere,
        a reference to that same object will be returned. Otherwise,
        a new object will be created.

        Parameters
        ----------
        scan_id : str
            The unique scan id value to be retrieved

        Returns
        -------
        Scan
        """
        if isinstance(scan_id, bytes):
            scan_id = scan_id.decode('utf8')
        try:
            return self._scan_cache[scan_id]
        except KeyError:
            packed = self._make_scan(self._get_scan_by_id_raw(scan_id))
            self._scan_cache[packed.id] = packed
            return packed

    def _get_scan_by_id_raw(self, scan_id):
        try:
            return self._source.get_by_id(scan_id)
        except AttributeError as exc:
            # If the source is somehow None, then we should not try to convert
            # the error into a KeyError
            if self._source is None:
                raise exc
            # When index-driven lookup fails, pyteomics will parse the XML
            # file from the beginning until it runs out of file
            err = KeyError(scan_id)
            raise err

    def get_scan_by_time(self, time):
        """Retrieve the scan object for the specified scan time.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        time : float
            The time to get the nearest scan from

        Returns
        -------
        Scan
        """
        scan_ids = tuple(self.index)
        lo = 0
        hi = len(scan_ids)

        best_match = None
        best_error = float('inf')

        if time == float('inf'):
            return self.get_scan_by_id(scan_ids[-1])

        while hi != lo:
            mid = (hi + lo) // 2
            sid = scan_ids[mid]
            scan = self.get_scan_by_id(sid)
            if not self._validate(scan):
                sid = scan_ids[mid - 1]
                scan = self.get_scan_by_id(sid)
                if not self._validate(scan):
                    sid = scan_ids[mid - 2]
                    scan = self.get_scan_by_id(sid)

            scan_time = scan.scan_time
            err = abs(scan_time - time)
            if err < best_error:
                best_error = err
                best_match = scan
            if scan_time == time:
                return scan
            elif (hi - lo) == 1:
                return best_match
            elif scan_time > time:
                hi = mid
            else:
                lo = mid
        if hi == 0 and not self._use_index:
            raise TypeError("This method requires the index. Please pass `use_index=True` during initialization")

    def get_scan_by_index(self, index):
        """Retrieve the scan object for the specified scan index.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        index: int
            The index to get the scan for

        Returns
        -------
        Scan
        """
        if not self._use_index:
            raise TypeError("This method requires the index. Please pass `use_index=True` during initialization")
        index_keys = tuple(self.index)
        id_str = index_keys[index]
        scan = self.get_scan_by_id(id_str)
        if not self._validate(scan):
            warnings.warn("index %d, id=%r does not appear to be a mass spectrum. Most behaviors will fail." % (
                index, id_str), stacklevel=2)
        return scan

    def _yield_from_index(self, scan_source, start):
        raise NotImplementedError()

    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        '''Reconstruct an iterator which will start from the scan matching one of ``scan_id``,
        ``rt``, or ``index``. Only one may be provided.

        After invoking this method, the iterator this object wraps will be changed to begin
        yielding scan bunchs (or single scans if ``grouped`` is ``False``).

        This method will trigger several random-access operations, making it prohibitively
        expensive for normally compressed files.

        Arguments
        ---------
        scan_id: str, optional
            Start from the scan with the specified id.
        rt: float, optional
            Start from the scan nearest to specified time (in minutes) in the run. If no
            exact match is found, the nearest scan time will be found, rounded up.
        index: int, optional
            Start from the scan with the specified index.
        require_ms1: bool, optional
            Whether the iterator must start from an MS1 scan. True by default.
        grouped: bool, optional
            whether the iterator should yield scan bunches or single scans. True by default.
        '''
        if scan_id is None:
            if rt is not None:
                scan = self.get_scan_by_time(rt)
            elif index is not None:
                try:
                    scan = self.get_scan_by_index(index)
                except IndexError:
                    if index > len(self.index):
                        index = len(self.index) - 1
                    else:
                        index = 0
                    scan = self.get_scan_by_index(index)

            else:
                raise ValueError("Must provide a scan locator, one of (scan_id, rt, index)")

            scan_id = scan.id
        else:
            scan = self.get_scan_by_id(scan_id)

        # We must start at an MS1 scan, so backtrack until we reach one
        if require_ms1:
            scan = self._locate_ms1_scan(scan)
            scan_id = scan.id

        iterator = self._yield_from_index(self._source, scan_id)
        self.make_iterator(iterator, grouped=grouped)
        return self

    def __repr__(self):
        return "{self.__class__.__name__}({self.source_file!r})".format(self=self) # pylint: disable=missing-format-attribute

    def __reduce__(self):
        return self.__class__, (self.source_file, self._use_index)


@xml._keepstate
def _find_section(source, section):
    value = next(source.iterfind(section))
    return value


@xml._keepstate
def get_tag_attributes(source, tag_name):
    '''Iteratively parse XML stream in ``source`` until encountering ``tag_name``
    at which point parsing terminates and return the attributes of the matched
    tag.

    Parameters
    ----------
    source: file-like
        A file-like object over an XML document
    tag_name: str
        The name of the XML tag to parse until

    Returns
    -------
    dict
    '''
    g = etree.iterparse(source, ('start', 'end'))
    for event, tag in g:
        if event == 'start':
            if xml._local_name(tag) == tag_name:
                return tag.attrib
            else:
                continue
        else:
            tag.clear()
    return None


@xml._keepstate
def iterparse_until(source, target_name, quit_name):
    '''Iteratively parse XML stream in ``source``, yielding XML elements
    matching ``target_name``. If at any point a tag matching ``quit_name``
    is encountered, stop parsing.

    Parameters
    ----------
    source: file-like
        A file-like object over an XML document
    tag_name: str
        The name of the XML tag to parse until
    quit_name: str
        The name to stop parsing at.

    Yields
    ------
    lxml.etree.Element
    '''
    g = etree.iterparse(source, ('start', 'end'))
    for event, tag in g:
        if event == 'start':
            if xml._local_name(tag) == quit_name:
                break
            else:
                if xml._local_name(tag) == target_name:
                    yield tag
                else:
                    tag.clear()



import unittest


from ms_deisotope.data_source import infer_type
from .common import datafile


class TestInferType(unittest.TestCase):
    mzml_path = datafile("three_test_scans.mzML")
    mzmlgz_path = datafile("three_test_scans.mzML.gz")

    def test_infer_loader(self):
        reader = infer_type.MSFileLoader(self.mzml_path)
        self.assertIsNotNone(next(reader))

    def test_guess_from_path(self):
        loader_t = infer_type.guess_type_from_path(self.mzml_path)
        self.assertEqual(loader_t, infer_type.MzMLLoader)

    def test_guess_from_file_sniffing(self):
        loader_t = infer_type.guess_type_from_file_sniffing(self.mzml_path)
        self.assertEqual(loader_t, infer_type.MzMLLoader)

    def test_infer_loader_compressed(self):
        reader = infer_type.MSFileLoader(self.mzmlgz_path)
        self.assertIsNotNone(next(reader))



'''Represent the basic structures of a mass spectrum and its processed contents,
and provide an interface for manipulating that data.
'''
import warnings

from collections import namedtuple

try:
    from collections import Sequence as _SequenceABC
except ImportError:
    from collections.abc import Sequence as _SequenceABC

from numbers import Number

import numpy as np

from ms_peak_picker import average_signal
from ms_peak_picker.base import PeakLike

from ms_deisotope.averagine import neutral_mass, mass_charge_ratio
from ms_deisotope.deconvolution import deconvolute_peaks

try:
    from ms_deisotope.plot import annotate_scan as _annotate_precursors, draw_raw
except ImportError:
    def _missing_matplotlib(*args, **kwargs):
        raise ImportError(
            "This method requires matplotlib. Please install it.")
    _annotate_precursors = _missing_matplotlib
    draw_raw = _missing_matplotlib

from ms_deisotope.utils import Constant


DEFAULT_CHARGE_WHEN_NOT_RESOLVED = 1
ChargeNotProvided = Constant("ChargeNotProvided")


class ScanBunch(namedtuple("ScanBunch", ["precursor", "products"])):
    """Represents a single MS1 scan and all MSn scans derived from it,
    or a collection of related MSn scans.

    Attributes
    ----------
    precursor: :class:`~.ScanBase`
        A single MS1 scan which may have undergone MSn
    products: list
        A list of 0 or more :class:`~.ScanBase` objects which were derived
        from :attr:`precursor` or another element of this list derived
        from it.
    """

    def __new__(cls, *args, **kwargs):
        inst = super(ScanBunch, cls).__new__(cls, *args, **kwargs)
        inst._id_map = {}
        if inst.precursor is not None:
            inst._id_map[inst.precursor.id] = inst.precursor
        for scan in inst.products:
            inst._id_map[scan.id] = scan
        return inst

    def precursor_for(self, scan):
        """Find the precursor :class:`~.ScanBase` instance
        for the given scan object

        Parameters
        ----------
        scan : :class:`~.ScanBase`
            The MSn scan to look for the MSn-1 scan for

        Returns
        -------
        :class:`~.ScanBase`
        """
        if scan.precursor_information is not None:
            scan_id = scan.precursor_information.precursor_scan_id
            return self.get_scan_by_id(scan_id)
        return None

    def get_scan_by_id(self, scan_id):
        """Retrieve the scan object for the specified scan id from this
        group in memory.

        Parameters
        ----------
        scan_id : str
            The unique scan id value to be retrieved

        Returns
        -------
        :class:`~.ScanBase`
        """
        return self._id_map[scan_id]

    def annotate_precursors(self, nperrow=4, ax=None):
        '''Plot the spectra in this group as a grid, with the full
        MS1 spectrum in profile in the top row, and each MSn spectrum's
        precursor ion revealed in a grid panel below, with isolation
        window and selected ion/monoisotopic peak annotated.

        Parameters
        ----------
        nperrow: :class:`int`
            The number of precursors to annotate per row
            in the grid.
        ax: :class:`matplotlib._axes.Axes`, optional
            The axis to draw on. If not provided, a new figure
            will be created, along with a new axis.

        Returns
        -------
        :class:`matplotlib._axes.Axes`
        '''
        return _annotate_precursors(
            self.precursor, self.products, nperrow=nperrow, ax=ax)

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        if cycle:
            p.text("ScanBunch(...)")
            return
        p.text("ScanBunch(\n")
        with p.group(2):
            with p.group(4, "precursor=\n"):
                p.pretty(self.precursor)
            with p.group(4, ",\nproducts=\n"):
                p.pretty(self.products)
        p.text(")")

    def pack(self):
        '''Build a new :class:`ScanBunch` where each scan in it is returned by calling
        :meth:`~.Scan.pack`

        Returns
        -------
        :class:`ScanBunch`
        '''
        return self.__class__(self.precursor.pack(), [
            p.pack() for p in self.products
        ])


class RawDataArrays(namedtuple("RawDataArrays", ['mz', 'intensity'])):
    """Represent the m/z and intensity arrays associated with a raw
    mass spectrum.

    Supports scaling and summing, as well as low level m/z search.

    Thin wrapper around a ``namedtuple``, so this object supports
    the same interfaces as a tuple.

    Attributes
    ----------
    mz: np.ndarray
        The m/z axis of a mass spectrum
    intensity: np.ndarray
        The intensity measured at the corresponding m/z of a mass spectrum
    """

    def __new__(cls, mz, intensity, arrays=None):
        inst = super(RawDataArrays, cls).__new__(cls, mz, intensity)
        inst.data_arrays = dict()
        if arrays:
            inst.data_arrays.update(arrays)
        return inst

    def __copy__(self):
        inst = self.__class__(self.mz.copy(), self.intensity.copy(), {
            k: v.copy() for k, v in self.data_arrays.items()
        })
        return inst

    def copy(self):
        """Make a deep copy of this object.

        Returns
        -------
        :class:`RawDataArray`
        """
        return self.__copy__()

    def plot(self, *args, **kwargs):
        """Draw the profile spectrum described by the
        contained arrays.

        Parameters
        ----------
        ax: :class:`matplotlib._axes.Axes`
            The figure axes onto which to draw the plot. If not provided,
            this will default to the current figure interactively.
        **kwargs
            All keywords are forwarded to :meth:`plot` on ``ax``.

        Returns
        -------
        :class:`matplotlib._axes.Axes`
            The axes drawn on
        """
        ax = draw_raw(self, *args, **kwargs)
        return ax

    def __eq__(self, other):
        try:
            return np.allclose(
                self[0], other[0]) and np.allclose(
                    self[1], other[1])
        except ValueError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __mul__(self, i):
        return self.__class__(self.mz, self.intensity * i)

    def __div__(self, d):
        return self.__class__(self.mz, self.intensity / d)

    def __add__(self, other):
        if len(self.mz) == len(other.mz) and np.allclose(self.mz, other.mz):
            return self.__class__(self.mz, self.intensity + other.intensity)
        else:
            return self.__class__(*average_signal([self, other])) * 2

    def find_mz(self, mz):
        """Find the nearest index to the query ``mz``

        Parameters
        ----------
        mz : float
            The m/z value to search for

        Returns
        -------
        int
            The index nearest to the query m/z
        """
        n = len(self.mz)
        lo = 0
        hi = n

        while hi != lo:
            mid = int((hi + lo) // 2)
            y = self.mz[mid]
            err = y - mz
            if abs(err) < 0.1:
                best_index = mid
                best_err = abs(err)
                i = mid
                while i >= 0:
                    y = self.mz[i]
                    err = y - mz
                    if err <= -0.1:
                        break
                    abs_err = abs(err)
                    if abs_err < best_err:
                        best_err = abs_err
                        best_index = i
                    i -= 1
                i = mid
                while i < n:
                    y = self.mz[i]
                    err = y - mz
                    if err >= 0.1:
                        break
                    abs_err = abs(err)
                    if abs_err < best_err:
                        best_err = abs_err
                        best_index = i
                    i += 1
                return best_index
            elif hi - lo == 1:
                return mid
            elif err > 0:
                hi = mid
            else:
                lo = mid
        return 0

    def between_mz(self, low, high):
        """Returns a slice of the arrays between ``low`` and ``high``
        m/z

        Parameters
        ----------
        low : float
            The lower bound m/z
        high : float
            The upper bound m/z

        Returns
        -------
        :class:`.RawDataArrays`
        """
        i = self.find_mz(low)
        j = self.find_mz(high) + 1
        if not (low <= self.mz[i] <= high):
            i += 1
        return self.__class__(self.mz[i:j], self.intensity[i:j])

    def __getitem__(self, i):
        if isinstance(i, int):
            return super(RawDataArrays, self).__getitem__(i)
        else:
            return self.arrays[i]


class ScanBase(object):
    '''Abstract base class for Scan-like objects
    '''

    def has_ion_mobility(self):
        '''Check whether this scan has drift time information associated with
        it.

        If this scan has been aggregated, it will only check the first scan in
        the aggregate.
        '''
        acq = self.acquisition_information
        if acq is None:
            return False
        scan_event = acq[0]
        return scan_event.has_ion_mobility()

    @property
    def drift_time(self):
        '''A convenience method to access the first
        scan event to retrieve its drift time.

        Returns
        -------
        float or None
        '''
        acq = self.acquisition_information
        if acq is None:
            return None
        scan_event = acq[0]
        return scan_event.drift_time

    @property
    def scan_id(self):
        '''An alias for :attr:`id`
        '''
        return self.id

    @property
    def tic(self):
        """A facade function for calculating the total ion current (TIC) of a spectrum.

        This exposes a facade object of type :class:`TICMethods` to take care of the different
        ways in which the TIC may be calculated.

        Returns
        -------
        :class:`TICMethods`

        Examples
        --------
        Just directly calling the `tic` attribute will use the most refined data source
        to calculate the TIC. This means that if the TIC is recalculated after refinement,
        the number may change.

        >>> from ms_deisotope.test.common import example_scan_bunch
        >>> bunch = example_scan_bunch()
        >>> bunch.precursor.tic()
        8886549.0
        >>> bunch.precursor.tic.raw()
        8886549.0

        The picked peaks can be used through :meth:`TICMethods.centroided`, which take
        priority over the raw signal when calling :meth:`tic` directly.
        >>> bunch.precursor.pick_peaks()
        >>> bunch.precursor.tic.centroided()
        8886548.890350103

        The deconvoluted peaks can be used through :meth:`TICMethods.deconvoluted`.
        >>> bunch.precursor.deconvolute(use_quick_charge=True)
        >>> bunch.precursor.tic.deconvoluted()
        8195619.241884331
        >>> bunch.precursor.tic()
        8195619.241884331

        """
        return TICMethods(self)

    @property
    def base_peak(self):
        """A facade function for calculating the base peak, the most abundant peak,
        of a spectrum.

        This exposes a facade object of type :class:`BasePeakMethods` to take care of
        the different ways in which the base peak may be calculated. The interface of
        this object is the same as the interface exposed by the :attr:`tic` attribute,
        but instead of returning a scalar float, it returns a :class:`~.PeakLike` object.

        Returns
        -------
        :class:`BasePeakMethods`

        See Also
        --------
        :attr:`tic`
        """
        return BasePeakMethods(self)

    @property
    def plot(self):
        """A facade function for plotting different layers of signal in the scan from raw,
        centroided, and deconvoluted representation, as well as limited precursor annotation.

        Returns
        -------
        :class:`PlottingMethods`
        """
        return PlottingMethods(self)

    @property
    def peaks(self):
        return PeakSetMethods(self)

    def copy(self, deep=True):
        """Return a deep copy of the :class:`Scan` object
        wrapping the same reference data.

        Returns
        -------
        :class:`ScanBase`
        """
        return self.clone(deep)

    def __copy__(self):
        return self.clone()

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, ScanBase):
            return False
        try:
            eq = (self.scan_id == other.scan_id) and (
                abs(self.scan_time - other.scan_time) < 1e-3) and (
                    self.index == other.index) and (
                        self.ms_level == other.ms_level)
            if not eq:
                return False
        except AttributeError:
            return False
        try:
            eq = self.arrays == other.arrays
            if not eq:
                return False
        except AttributeError:
            # ProcessedScan doesn't have an arrays attribute
            pass
        try:
            eq = self.peak_set == other.peak_set
            if not eq:
                return False
        except AttributeError:
            if ((self.peak_set is None and other.peak_set is not None) or (
                    self.peak_set is not None and other.peak_set is None)):
                pass
            else:
                return False

        try:
            eq = self.deconvoluted_peak_set == other.deconvoluted_peak_set
            if not eq:
                return False
        except AttributeError:
            if ((self.deconvoluted_peak_set is None and other.deconvoluted_peak_set is not None) or (
                    self.deconvoluted_peak_set is not None and other.deconvoluted_peak_set is None)):
                pass
            else:
                return False

        eq = self.precursor_information == other.precursor_information
        if not eq:
            return False
        eq = self.isolation_window == other.isolation_window
        if not eq:
            return False
        try:
            a = self.acquisition_information
            b = other.acquisition_information
            if a is not None and b is not None:
                eq = a == b
            else:
                eq = True
            if not eq:
                return False
        except AttributeError:
            pass
        try:
            a = self.activation
            b = other.activation
            if a is not None and b is not None:
                eq = a == b
            else:
                eq = True
            if not eq:
                return False
        except AttributeError:
            pass

        return True

    def __ne__(self, other):
        return not (self == other)

    def bind(self, source):
        '''Attach this object and its other referent members
        to ``source``, letting them load information.
        '''
        if self.precursor_information is not None:
            self.precursor_information.bind(source)
        return self

    def unbind(self):
        '''Detattch this object and its other referent members
        from their currently bound :attr:`source`.

        This may cause errors if more information is requested but is not
        cached, or if requesting another :class:`ScanBase` be loaded.
        '''
        if self.precursor_information is not None:
            self.precursor_information.unbind()
        return self


class PrecursorInformation(object):
    """Store information relating a tandem MS scan to its precursor MS scan.

    .. note::
        The attributes prefixed with `extracted_` refer to the quantities estimated
        from the data, while those unprefixed are the values read directly from the
        data source. These values regularly do not agree. When available, the extracted
        values should be more accurate.

    Attributes
    ----------
    charge : int
        The charge reported in the source metadata
    defaulted : bool
        Whether the information in the extracted fields reflects empirical
        information or fell back on the vendor-reported values.
    extracted_charge : int
        The charge estimated from the source data
    extracted_intensity : float
        The sum of the peak heights of the extracted isotopic pattern
    extracted_neutral_mass : float
        The monoisotopic neutral mass estimated from the source data
    extracted_peak : :class:`.DeconvolutedPeak`
        The deconvoluted peak summarizing the precursor ion
    intensity : float
        The abundance reported in the source metadata
    mz : float
        The m/z reported in the source metadata
    orphan : bool
        Whether there was an isotopic pattern to extract in the precursor scan. Usually
        paired with `defaulted`
    peak : :class:`.FittedPeak`
        The peak nearest :attr:`mz`, and the starting point for estimating information
        about the precursor ion
    precursor_scan_id : str
        The id string for the precursor scan
    source : :class:`ScanIterator`
        Any object implementing the :class:`ScanIterator` interface to be used to look up
        the precursor scan with :attr:`precursor_scan_id`
    """

    def __init__(self, mz, intensity, charge, precursor_scan_id=None, source=None,
                 extracted_neutral_mass=0, extracted_charge=0, extracted_intensity=0,
                 peak=None, extracted_peak=None, defaulted=False, orphan=False,
                 product_scan_id=None, annotations=None, coisolation=None):
        try:
            charge = int(charge)
        except Exception:
            pass
        try:
            extracted_charge = int(extracted_charge)
        except Exception:
            pass
        if not annotations:
            annotations = {}
        if not coisolation:
            coisolation = []

        self.mz = mz
        self.intensity = intensity
        self.charge = charge

        self.precursor_scan_id = precursor_scan_id
        self.source = source

        self.extracted_neutral_mass = extracted_neutral_mass
        self.extracted_charge = extracted_charge
        self.extracted_intensity = extracted_intensity

        self.peak = peak
        self.extracted_peak = extracted_peak
        self.defaulted = defaulted
        self.orphan = orphan
        self.product_scan_id = product_scan_id

        self.annotations = annotations
        self.coisolation = coisolation

    def __repr__(self):
        return "PrecursorInformation(mz=%0.4f/%0.4f, intensity=%0.4f/%0.4f, charge=%r/%r, scan_id=%r)" % (
            self.mz,
            self.extracted_mz if self.extracted_neutral_mass != 0. else 0.,
            self.intensity or 0., self.extracted_intensity or 0., self.charge,
            self.extracted_charge or 0., self.precursor_scan_id)

    def __reduce__(self):
        return self.__class__, (0, 0, 0), self.__getstate__()

    def __getstate__(self):
        # explicitly do not propagate :attr:`source` when serializing.
        return (self.mz, self.intensity, self.charge, self.precursor_scan_id, None, self.extracted_neutral_mass,
                self.extracted_charge, self.extracted_intensity, self.peak, self.extracted_peak,
                self.defaulted, self.orphan, self.product_scan_id, self.annotations, self.coisolation)

    def __setstate__(self, state):
        (self.mz, self.intensity, self.charge, self.precursor_scan_id, self.source, self.extracted_neutral_mass,
         self.extracted_charge, self.extracted_intensity, self.peak, self.extracted_peak,
         self.defaulted, self.orphan, self.product_scan_id) = state[:13]
        if len(state) > 13:
            self.annotations = state[13]
        if len(state) > 14:
            self.coisolation = list(state[14])

    def __eq__(self, other):
        if other is None:
            return False
        eq = self.precursor_scan_id == other.precursor_scan_id
        if not eq:
            return False
        eq = self.product_scan_id == other.product_scan_id
        if not eq:
            return False
        self_fit = self.extracted_neutral_mass != 0
        other_fit = other.extracted_neutral_mass != 0
        self_mass = self.extracted_mz if self_fit else self.mz
        other_mass = other.extracted_mz if other_fit else other.mz
        eq = np.isclose(self_mass, other_mass)
        if not eq:
            return False
        self_charge = self.extracted_charge if self_fit else self.charge
        other_charge = other.extracted_charge if other_fit else other.charge
        eq = self_charge == other_charge
        if not eq:
            return False
        return True

    def __ne__(self, other):
        return not (self == other)

    def bind(self, source):
        '''Attach this object and its other referent members
        to ``source``, letting them load information.
        '''
        self.source = source
        return self

    def unbind(self):
        '''Detattch this object the currently bound :attr:`source`.
        '''
        self.source = None

    def extract(self, peak, override_charge=None):
        '''Populate the extracted attributes of this object from the attributes
        of a :class:`~.DeconvolutedPeak` instance.

        Parameters
        ----------
        peak: :class:`~.DeconvolutedPeak`
            The peak to copy attributes from
        override_charge: :class:`int`, optional
            If provided, this charge will be used instead of the charge of ``peak``
        '''
        self.extracted_neutral_mass = peak.neutral_mass
        self.extracted_charge = int(
            peak.charge) if override_charge is None else override_charge
        self.extracted_intensity = peak.intensity
        self.extracted_peak = peak
        self.orphan = False
        self.defaulted = False

    def default(self, orphan=False):
        '''Populate the extracted attributes of this object from the matching
        original attributes.

        This usually reflects a failure to find an acceptable deconvolution solution,
        and may indicate that there was no peak at the specified location when ``orphan``
        is :const:`True`

        Parameters
        ----------
        orphan: :class:`bool`
            Whether or not to set :attr:`orphan` to :const:`True`, indicating no peak was
            found near :attr:`mz`.
        '''
        if self.charge == ChargeNotProvided:
            warnings.warn(
                "A precursor has been defaulted with an unknown charge state.")
            self.extracted_charge = ChargeNotProvided
            self.extracted_neutral_mass = neutral_mass(self.mz, DEFAULT_CHARGE_WHEN_NOT_RESOLVED)
        else:
            self.extracted_charge = int(self.charge)
            self.extracted_neutral_mass = self.neutral_mass
        self.extracted_intensity = self.intensity
        self.defaulted = True
        if orphan:
            self.orphan = True

    @property
    def neutral_mass(self):
        """Calculate the neutral mass of the precursor from the given m/z and charge.

        Returns
        -------
        float
        """
        if self.charge == ChargeNotProvided:
            warnings.warn(
                "A precursor with an unknown charge state was used to compute a neutral mass.")
            return neutral_mass(self.mz, DEFAULT_CHARGE_WHEN_NOT_RESOLVED)
        return neutral_mass(self.mz, self.charge)

    @property
    def extracted_mz(self):
        """Recalculate the m/z of the precursor from the fitted neutral mass and charge

        Returns
        -------
        float
        """
        if self.extracted_charge == ChargeNotProvided or (
                self.extracted_charge == 0 and self.charge == ChargeNotProvided):
            warnings.warn(
                "A precursor with an unknown charge state was used to compute a m/z.")
            return mass_charge_ratio(self.mz, DEFAULT_CHARGE_WHEN_NOT_RESOLVED)
        return mass_charge_ratio(self.extracted_neutral_mass, self.extracted_charge)

    @property
    def precursor(self):
        """The scan in which the precursor ion was isolated.

        Returns
        -------
        :class:`ScanBase`
        """
        if self.precursor_scan_id is None:
            return None
        return self.source.get_scan_by_id(self.precursor_scan_id)

    @property
    def product(self):
        """The scan in which the precursor ion was fragmented and daughter ions were observed.

        Returns
        -------
        :class:`ScanBase`
        """
        if self.product_scan_id is None:
            return None
        return self.source.get_scan_by_id(self.product_scan_id)

    def copy(self):
        """Make a shallow copy of this object.

        Returns
        -------
        :class:`PrecursorInformation`
        """
        dup = self.__class__(
            self.mz, self.intensity, self.charge, self.precursor_scan_id, self.source,
            self.extracted_neutral_mass, self.extracted_charge, self.extracted_intensity,
            self.peak, self.extracted_peak, self.defaulted, self.orphan,
            self.product_scan_id, self.annotations, self.coisolation)
        return dup

    def clone(self):
        """Make a shallow copy of this object.

        .. note::

            This is an alias of :meth:`copy`

        Returns
        -------
        :class:`PrecursorInformation`
        """
        return self.copy()

    def correct_mz(self, error_tolerance=2e-5, enforce_isolation_window=False):
        """Find the peak nearest to :attr:`mz` in :attr:`precursor` and
        update :attr:`mz` from it.

        .. note::
            The peak selected may still not be the monoisotopic peak. This requires
            a deconvolution procedure.

        Parameters
        ----------
        error_tolerance: float, optional
            The error tolerance in PPM to use when searching for the nearest peak (the default is 2e-5).
        enforce_isolation_window: bool, optional
            Whether or not to force the specified m/z. Defaults to :const:`False`.
        """
        if self.precursor_scan_id is None:
            return
        precursor_scan = self.precursor
        if precursor_scan is None:
            return
        if precursor_scan.peak_set is None:
            precursor_scan.pick_peaks()
        peaks = precursor_scan.peak_set
        peak = peaks.has_peak(self.mz, error_tolerance)
        if peak is not None:
            self.mz = peak.mz
        if enforce_isolation_window and self.product_scan_id is not None:
            product_scan = self.product
            if product_scan is not None:
                isolation_window = product_scan.isolation_window
            else:
                isolation_window = None
            if isolation_window is not None and not isolation_window.is_empty():
                if not isolation_window.spans(self.mz):
                    region = peaks.between(
                        isolation_window.lower_bound, isolation_window.upper_bound)
                    if region:
                        peak = max(region, key=lambda x: x.intensity)
                        self.mz = peak.mz

    def find_monoisotopic_peak(self, trust_charge_state=True, precursor_scan=None, **kwargs):
        """Find the monoisotopic peak for this precursor.

        This convenience method carries out a simplified procedure for finding the
        precursor ion's monoisotpic peak and charge state in the precursor scan. It
        follows steps similar to those found in the :class:`~.ScanProcessor` pipeline,
        but is not as flexible or complete.

        .. note::

            For a full deconvolution result, please use :class:`~.ScanProcessor`, which
            carries out a more complete error checking procedure.


        Parameters
        ----------
        trust_charge_state: bool, optional
            Whether or not to trust the original precursor charge state, which may be based
            upon information not available in the examined mass spectrum.
        precursor_scan: :class:`~.ScanBase`, optional
            The spectrum to look for the precursor peak in. If not provided,
            :attr:`precursor` will be used.

        Returns
        -------
        :class:`float`:
            The updated m/z of the precursor ion's monoisotopic peak.
        :class:`bool`:
            Whether or not the deconvolution procedure was able to run successfully.

        """
        if precursor_scan is None:
            if self.precursor_scan_id is None:
                return False
            precursor_scan = self.precursor
            if precursor_scan is None:
                return False
        if precursor_scan.peak_set is None:
            precursor_scan.pick_peaks()
        charge_range = kwargs.get("charge_range", (1, 8))
        if precursor_scan.polarity < 0 and max(charge_range) > 0:
            charge_range = tuple(c * precursor_scan.polarity for c in charge_range)
        kwargs['charge_range'] = charge_range

        # Obtain the peaks around the expected precursor peak
        peaks = precursor_scan.peak_set.between(self.mz - 3, self.mz + 6)
        peaks = peaks.clone()
        peaks.reindex()
        ref_peak = peaks.has_peak(self.mz, 2e-5)

        # No experimental peak found, so mark that this precursor is an orphan and default it
        if ref_peak is None:
            self.default(orphan=True)
            return self.extracted_mz, False

        priority_target = [ref_peak]

        result = deconvolute_peaks(peaks, priority_list=priority_target, **kwargs)
        _, priority_results = result
        peak = priority_results[0]

        # No deconvoluted peak found, so mark that this precursor is an orphan and default it
        if peak is None:
            self.default(orphan=True)
            return self.extracted_mz, False

        # A peak was found, we don't know the expected charge state, so accept it
        # and extract the updated peak fit
        elif self.charge == ChargeNotProvided:
            self.extract(peak)
            return self.extracted_mz, True

        # A peak was found, but it doesn't match the trusted charge state, so reject
        # it and default this precursor
        if trust_charge_state and self.charge != peak.charge:
            self.default()
            return self.extracted_mz, False
        # The returned peak matches the expected charge state, so accept it and
        # extract the updated peak fit
        self.extract(peak)
        return self.extracted_mz, True


class TICMethods(object):
    """A helper class that will figure out the most refined signal source to
    calculate the total ion current from.
    """
    def __init__(self, scan):
        self.scan = scan

    def _peak_sequence_tic(self, peaks):
        total = 0
        for peak in peaks:
            total += peak.intensity
        return total

    def _simple_tic(self, points):
        return sum(points)

    def _tic_raw_data_arrays(self, arrays):
        return arrays.intensity.sum()

    def __call__(self):
        return self._guess()

    def _guess(self):
        """Guess which strategy to use to calculate the most refined representation of
        the TIC.

        Returns
        -------
        float
        """
        try:
            return self.deconvoluted()
        except (AttributeError, TypeError):
            pass
        try:
            return self.centroided()
        except (AttributeError, TypeError):
            pass
        try:
            return self.raw()
        except (AttributeError, TypeError):
            pass

        points = list(self.scan)
        if points:
            # This may not work if PeakLike is recognizing an external peak-like object
            # that is not derived from PeakBase and C-extensions are enabled?
            if isinstance(points[0], PeakLike):
                return self._peak_sequence_tic(points)
            elif isinstance(points[0], Number):
                return self._simple_tic(points)
            else:
                raise TypeError(
                    "Cannot determine how to calculate a TIC from %r of type %r" % (
                        self.scan, type(self.scan)))
        else:
            raise TypeError(
                "Cannot determine how to calculate a TIC from %r of type %r" % (
                    self.scan, type(self.scan)))

    def raw(self):
        """Calculate the TIC from the raw intensity signal of the spectrum with no processing.

        Returns
        -------
        float
        """
        return self._tic_raw_data_arrays(self.scan.arrays)

    def centroided(self):
        """Calculate the TIC from the picked peak list of the spectrum.

        Returns
        -------
        float
        """
        return self._peak_sequence_tic(self.scan.peak_set)

    def deconvoluted(self):
        """Calculate the TIC from the deconvoluted peak list of the spectrum.

        Returns
        -------
        float
        """
        return self._peak_sequence_tic(self.scan.deconvoluted_peak_set)


class BasePeakMethods(object):
    """A helper class that will figure out the most refined signal source to
    calculate the base peak from.
    """

    base_peak_t = namedtuple("BasePeak", ("mz", "intensity"))

    def __init__(self, scan):
        self.scan = scan

    def _peak_sequence_bp(self, peaks):
        if not peaks:
            return None
        peak = max(peaks, key=lambda x: x.intensity)
        return peak

    def _bp_raw_data_arrays(self, arrays):
        i = np.argmax(arrays.intensity)
        return self.base_peak_t(arrays.mz[i], arrays.intensity[i])

    def __call__(self):
        return self._guess()

    def _guess(self):
        """Guess which strategy to use to produce the most refined representation
        of the base peak.

        Returns
        -------
        :class:`~.PeakLike`
        """
        try:
            return self.deconvoluted()
        except (AttributeError, TypeError):
            pass
        try:
            return self.centroided()
        except (AttributeError, TypeError):
            pass
        try:
            return self.raw()
        except (AttributeError, TypeError):
            pass

        points = list(self.scan)
        if points:
            if isinstance(points[0], PeakLike):
                return self._peak_sequence_bp(points)
            raise TypeError(
                "Cannot determine how to calculate a base peak from %r of type %r" % (
                    self.scan, type(self.scan)))
        else:
            raise TypeError(
                "Cannot determine how to calculate a base peak from %r of type %r" % (
                    self.scan, type(self.scan)))

    def raw(self):
        """Calculate the base peak from the raw intensity signal of the spectrum with no processing.

        Returns
        -------
        :class:`~.PeakLike`
        """
        return self._bp_raw_data_arrays(self.scan.arrays)

    def centroided(self):
        """Calculate the base peak from the picked peak list of the spectrum.

        Returns
        -------
        :class:`~.FittedPeak`
        """
        return self._peak_sequence_bp(self.scan.peak_set)

    def deconvoluted(self):
        """Calculate the base peak from the deconvoluted peak list of the spectrum.

        Returns
        -------
        :class:`~.DeconvolutedPeak`
        """
        return self._peak_sequence_bp(self.scan.deconvoluted_peak_set)


class PeakSetMethods(_SequenceABC):
    """A facade to simplify determining how to call common peak-set methods
    on an object like :class:`~.Scan` which may have multiple tiers of
    representation of the peaks it contains, while still supporting directly
    using real PeakSet-like objects.

    In addition to the three common peak searching methods, this type supports
    :class:`~.Sequence` operations.

    Attributes
    ----------
    scan: object
        The scan containing peaks to manipulate
    is_scan: bool
        Whether or not :attr:`scan` is an instance of :class:`~.ScanBase`
    is_deconvoluted: bool
        Whether or not :attr:`scan` has been deconvoluted.
    is_centroided: bool
        Whether or not :attr:`scan` has been centroided (undergone peak picking).
    is_raw_sequence: bool
        Whether or not :attr:`scan` is of a known type or just a sequence.

    """
    def __init__(self, scan):
        self.scan = scan
        self.is_scan = self._is_scan()
        self.is_deconvoluted = self._is_deconvoluted()
        self.is_centroided = self._is_centroided()
        self.is_raw_sequence = self._is_raw_sequence()

    def _is_scan(self):
        return isinstance(self.scan, ScanBase)

    def _is_deconvoluted(self):
        if self.is_scan:
            return self.scan.deconvoluted_peak_set is not None
        else:
            from ms_deisotope.peak_set import DeconvolutedPeakSet
            return isinstance(self.scan, DeconvolutedPeakSet)

    def deconvoluted(self):
        """Get the deconvoluted peak set

        Returns
        -------
        :class:`~.DeconvolutedPeakSet`
        """
        if self.is_deconvoluted:
            return self._get_deconvoluted()
        return None

    def centroided(self):
        """Get the centroided peak set

        Returns
        -------
        :class:`~.PeakSet` or :class:`~.PeakIndex`
        """
        if self.is_centroided:
            return self._get_centroided()

    def raw(self):
        """Get the raw signal, or lacking that, the simplest representation available.

        Returns
        -------
        objects
        """
        if self.is_scan:
            return self.scan.arrays
        elif self.is_raw_sequence:
            return self.scan
        else:
            return None

    def _get_deconvoluted(self):
        if self.is_scan:
            return self.scan.deconvoluted_peak_set
        elif self.is_deconvoluted:
            return self.scan
        else:
            return None

    def _is_centroided(self):
        if self.is_scan:
            return self.scan.peak_set is not None
        else:
            from ms_peak_picker import PeakIndex, PeakSet
            return isinstance(self.scan, (PeakIndex, PeakSet))

    def _get_centroided(self):
        if self.is_scan:
            return self.scan.peak_set
        elif self.is_centroided:
            return self.scan
        else:
            return None

    def _is_raw_sequence(self):
        if self.is_scan:
            return False
        else:
            if hasattr(self.scan, 'has_peak'):
                return False
            return isinstance(self.scan, _SequenceABC)

    def has_peak(self, m, error_tolerance=2e-5):
        """Search the most refined representation available for a peak at the
        given mass dimension coordinates with the specified parts-per-million
        error tolerance.

        This method invokes the underlying collection's ``has_peak`` method,
        and may not expose all specialized parameters. If the underlying collection
        cannot support this operation, an error will be raised.

        Parameters
        ----------
        m : float
            The mass dimension coordinates to search for. Depending upon if the
            peaks have been deconvoluted, this may be m/z or neutral mass.
        error_tolerance : float, optional
            The parts-per-million error tolerance to apply (the default is 2e-5)

        Returns
        -------
        :class:`~.PeakBase`

        See Also
        --------
        :meth:`~.DeconvolutedPeakSet.has_peak`
        :meth:`~.PeakSet.has_peak`

        Raises
        ------
        NotImplementedError
            When the underlying collection does not support calling ``has_peak``
        """
        if self.is_deconvoluted:
            return self._get_deconvoluted().has_peak(m, error_tolerance)
        elif self.is_centroided:
            return self._get_centroided().has_peak(m, error_tolerance)
        elif self.is_scan:
            self.scan.pick_peaks()
            self.is_centroided = self._is_centroided()
            return self._get_centroided().has_peak(m, error_tolerance)
        elif self.is_raw_sequence:
            if not hasattr(self.scan, 'has_peak'):
                raise NotImplementedError()
            return self.scan.has_peak(m, error_tolerance)
        else:
            raise NotImplementedError()

    def all_peaks_for(self, m, error_tolerance=2e-5):
        """Search the most refined representation available for all peaks at the
        given mass dimension coordinates with the specified parts-per-million
        error tolerance.

        This method invokes the underlying collection's ``all_peaks_for`` method,
        and may not expose all specialized parameters. If the underlying collection
        cannot support this operation, an error will be raised.

        Parameters
        ----------
        m : float
            The mass dimension coordinates to search for. Depending upon if the
            peaks have been deconvoluted, this may be m/z or neutral mass.
        error_tolerance : float, optional
            The parts-per-million error tolerance to apply (the default is 2e-5)

        Returns
        -------
        :class:`tuple` of :class:`~.PeakBase`

        See Also
        --------
        :meth:`~.DeconvolutedPeakSet.all_peaks_for`
        :meth:`~.PeakSet.all_peaks_for`

        Raises
        ------
        NotImplementedError
            When the underlying collection does not support calling ``all_peaks_for``
        """
        if self.is_deconvoluted:
            return self._get_deconvoluted().all_peaks_for(m, error_tolerance)
        elif self.is_centroided:
            return self._get_centroided().all_peaks_for(m, error_tolerance)
        elif self.is_scan:
            self.scan.pick_peaks()
            self.is_centroided = self._is_centroided()
            return self._get_centroided().all_peaks_for(m, error_tolerance)
        elif self.is_raw_sequence:
            if not hasattr(self.scan, 'all_peaks_for'):
                raise NotImplementedError()
            return self.scan.all_peaks_for(m, error_tolerance)
        else:
            raise NotImplementedError()

    def between(self, lo, hi, **kwargs):
        """Search the most refined representation available for all peaks at the
        between the given low and high mass dimension coordinates.

        This method invokes the underlying collection's ``between`` method,
        and may not expose all specialized parameters. If the underlying collection
        cannot support this operation, an error will be raised.

        Parameters
        ----------
        lo : float
            The lower bound mass dimension coordinates to search for.
            Depending upon if the peaks have been deconvoluted, this may be m/z
            or neutral mass.
        hi : float
            The upper bound mass dimension coordinates to search for.
            Depending upon if the peaks have been deconvoluted, this may be m/z
            or neutral mass.
        **kwargs
            Forwarded to the underlying collection's ``between`` method.

        Returns
        -------
        object

        See Also
        --------
        :meth:`~.DeconvolutedPeakSet.between`
        :meth:`~.PeakSet.between`

        Raises
        ------
        NotImplementedError
            When the underlying collection does not support calling ``between``
        """
        if self.is_deconvoluted:
            return self._get_deconvoluted().between(lo, hi, **kwargs)
        elif self.is_centroided:
            return self._get_centroided().between(lo, hi, **kwargs)
        elif self.is_scan:
            self.scan.pick_peaks()
            self.is_centroided = self._is_centroided()
            return self._get_centroided().between(lo, hi, **kwargs)
        elif self.is_raw_sequence:
            if not hasattr(self.scan, 'between'):
                raise NotImplementedError()
            return self.scan.between(lo, hi, **kwargs)
        else:
            raise NotImplementedError()

    def __len__(self):
        if self.is_raw_sequence:
            return len(self.scan)
        elif self.is_deconvoluted:
            return len(self._get_deconvoluted())
        elif self.is_centroided:
            return len(self._get_centroided())
        elif self.is_scan:
            self.scan.pick_peaks()
            self.is_centroided = self._is_centroided()
            return len(self._get_centroided())
        elif self.is_raw_sequence:
            return len(self.scan)
        else:
            raise NotImplementedError()

    def __getitem__(self, i):
        if self.is_raw_sequence:
            return (self.scan)[i]
        elif self.is_deconvoluted:
            return (self._get_deconvoluted())[i]
        elif self.is_centroided:
            return (self._get_centroided())[i]
        elif self.is_scan:
            self.scan.pick_peaks()
            self.is_centroided = self._is_centroided()
            return (self._get_centroided())[i]
        elif self.is_raw_sequence:
            return self.scan[i]
        else:
            raise NotImplementedError()


class PlottingMethods(object):
    """A plotting method facade that knows how to draw different facets of a spectrum.

    When called directly, the behavior is the same as calling :meth:`raw`, :meth:`centroided`,
    and meth:`deconvoluted` with a shared ``ax`` argument and common ``**kwargs``.
    """
    def __init__(self, scan):
        self.scan = scan
        self._plot_api = None
        from ms_deisotope import plot
        self._plot_api = plot

    def raw(self, *args, **kwargs):
        """Draws un-centroided profile data, visualizing continuous
        data points

        Parameters
        ----------
        ax : :class:`matplotlib.Axes`, optional
            The axis to draw the plot on. If missing, a new one will be created using
            :func:`matplotlib.pyplot.subplots`
        pretty: bool, optional
            If `True`, will call :func:`_beautify_axes` on `ax`
        normalize: bool, optional
            if `True`, will normalize the abundance dimension to be between 0 and 100%
        **kwargs
            Passed to :meth:`matplotlib.Axes.plot`

        Returns
        -------
        :class:`~.Axes`

        See Also
        --------
        :func:`ms_deisotope.plot.draw_raw`
        """
        return self._plot_api.draw_raw(self.scan.arrays, *args, **kwargs)

    def centroided(self, *args, **kwargs):
        """Draws centroided peak data, visualizing peak apexes.

        If :attr:`scan.peak_set` is not populated, this will fail.

        Parameters
        ----------
        ax : matplotlib.Axes, optional
            The axis to draw the plot on. If missing, a new one will be created using
            :func:`matplotlib.pyplot.subplots`
        pretty: bool, optional
            If `True`, will call :func:`_beautify_axes` on `ax`
        normalize: bool, optional
            if `True`, will normalize the abundance dimension to be between 0 and 100%
        **kwargs
            Passed to :meth:`matplotlib.Axes.plot`

        Returns
        -------
        matplotlib.Axes

        See Also
        --------
        :func:`ms_deisotope.plot.draw_peaklist`
        """
        return self._plot_api.draw_peaklist(self.scan.peak_set, *args, **kwargs)

    def deconvoluted(self, *args, **kwargs):
        """Draws deconvoluted peak data, visualizing collapsed envelope monoisotopic peaks.

        If :attr:`scan.peak_set` is not populated, this will fail.

        Parameters
        ----------
        ax : matplotlib.Axes, optional
            The axis to draw the plot on. If missing, a new one will be created using
            :func:`matplotlib.pyplot.subplots`
        pretty: bool, optional
            If `True`, will call :func:`_beautify_axes` on `ax`
        normalize: bool, optional
            if `True`, will normalize the abundance dimension to be between 0 and 100%
        **kwargs
            Passed to :meth:`matplotlib.Axes.plot`

        Returns
        -------
        matplotlib.Axes

        See Also
        --------
        :func:`ms_deisotope.plot.draw_peaklist`
        """
        return self._plot_api.draw_peaklist(self.scan.deconvoluted_peak_set, *args, **kwargs)

    def annotate_precursor(self, *args, **kwargs):
        '''Draw a zoomed-in view of the precursor scan of :attr:`scan` surrounding the
        area around each precursor ion that gave rise to :attr:`scan`
        with monoisotopic peaks and isolation windows marked.

        Parameters
        ----------
        ax: :class:`matplotlib._axes.Axes`
            An :class:`~.Axes` object to draw the plot on

        Returns
        -------
        :class:`matplotlib._axes.Axes`

        See Also
        --------
        :func:`ms_deisotope.plot.annotate_scan_single`
        '''
        pinfo = self.scan.precursor_information
        if pinfo is None:
            return
        precursor = pinfo.precursor
        return self._plot_api.annotate_scan_single(precursor, self.scan, *args, **kwargs)

    def label_peaks(self, *args, **kwargs):
        return self._plot_api.label_peaks(self.scan, *args, **kwargs)

    def annotate_isotopic_peaks(self, *args, **kwargs):
        """Mark distinct isotopic peaks from the :class:`~.DeconvolutedPeakSet`
        in ``scan``.

        Parameters
        ----------
        color_cycle: :class:`~.Iterable`
            An iterable to draw isotopic cluster colors from
        ax: :class:`matplotlib._axes.Axes`
            An :class:`~.Axes` object to draw the plot on

        Returns
        -------
        :class:`matplotlib._axes.Axes`

        See Also
        --------
        :func:`ms_deisotope.plot.annotate_isotopic_peaks`
        """
        return self._plot_api.annotate_isotopic_peaks(self.scan, *args, **kwargs)

    def _guess(self, ax=None, **kwargs):
        try:
            if self.scan.arrays:
                ax = self.raw(ax=ax)
        except (AttributeError, TypeError):
            pass
        try:
            if self.scan.peak_set:
                ax = self.centroided(ax=ax)
        except (AttributeError, TypeError):
            pass
        try:
            if self.scan.deconvoluted_peak_set:
                ax = self.deconvoluted(ax=ax)
        except (AttributeError, TypeError):
            pass
        return ax

    def __call__(self, ax=None, **kwargs):
        return self._guess()


try:
    from ms_deisotope._c.utils import _peak_sequence_tic, _peak_sequence_bp
    TICMethods._peak_sequence_tic = _peak_sequence_tic
    BasePeakMethods._peak_sequence_bp = _peak_sequence_bp
except ImportError:
    pass



'''
This directory contains libraries and code for reading file formats
using vendorized dependencies, not necessarily for MS vendor binary
files. These vendorized libraries may still require other platform-dependent
resources such as installed DLLs or 3rd party programs.

Each reader has had it's author's license prepended.
'''
